/*
*Copyright(c)2014, Jeffrey Lee
*Allrightsreserved.
*
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
*/

#include "common.h"
#include "decgen.h"
#include "node.h"
#include "encoding.h"
#include "actiondef.h"
#include "opcodesubset.h"
#include "input.h"
#include "constraint.h"
#include "funcdef.h"

/* TODO:
 - HTMLify readme?
 - automatic inlining
   - if child node is only referenced by one parent, inline it
   - also analyse any common parameters/constraints so they can be extracted/computed at the parent's level
     - with parameter to control maximum to extract?
 - add -inline option?
   - or use compile-time macro(s) to construct function prototypes/signatures 
 - add support for assembler output?
 - add support for variable-length instructions
   - start by restricting mask & constraint checks to only those which examine the loaded bytes
   - then when we reach a subset which doesn't contain any short instructions, insert a 'node_expand' which loads the extra bytes, then continue the tree from there
   - use syntax -bits=16,32 to list all the instruction lengths used?
     - so lengths can still be checked for errors
 - >32bit instruction length
   - need to use appropriate data type
   - compile-time option to set data type used by decgen?
   - and change 'uint32_t' to 'decgen_uint' or something?
     - would have to assume that the user uses the correct width type (i.e. doesn't truncate opcode, or any computations performed by constraints)
 - local macros
   - macros defined within encoding files, e.g. as shorthand for custom/complex constraints?
   - and ability to define macros as a result of -D options
     - to make command line cleaner when dealing with [CC], or parameterised macros for ARM/Thumb VFP/ASIMD merge, etc.
     - allows function/form of macro to change without everyone having to update their command lines
 - tool to parse the encoding files and strip out encodings unsupported by certain architectures?
   - nah, do it all in the actiondef files
     - unless it's possible to split the files into multiple sections using a handful of UNDEFINED rules for unimplemented instructions in certain architectures?
 - use new ambiguity/completeness check as better way of handling -exact?
   - for any completeness checks that would result in a failure, produce a new encoding
   - could be a problem if the gap is near the left of the opcode though
 - upgrade to allow ambiguous encodings iff they're kin?
   - code practically supports it already, doesn't it?
 - add built in test code?
   - outputs tree C file with built in actions to verify correctness
   - would mainly be for checking the generator for bugs (e.g. switch()) 
 - add coverage test
   - verify that each node/leaf/edge does get used and therefore is relevant
   - because the "nomatch" constraints may get put together in such a way that it's possible to deny more checks/encodings than the current code knows how?
     - for any encoding in the subset, if one of the nomatch constraints matches all the time, that encoding can be discarded
     - also more complex cases, e.g. if encoding needs {lor(A,B)} and nomatch 1 is {A} and nomatch 2 is {B}
       - would need to evaluate each nomatch for all combinations to make sure that at least one valid opcode is left
 - add -skeleton option to generate skeleton action file from an encoding file
 - add 'smart node'?
   - if two encodings (== 2 kin) left, just pick the least complex set of constraints and evaluate them on the opcode to determine which one it is
   - would save pointless experiments being produced that lead to needlessly deep trees
   - however would have to be careful that tree optimisation doesn't kill it
     - if constraints which are already known to be true were automatically eliminated from the check
 - big endian support?
   - invert bit order
 - improve child promotion code
   - it's possible that it could create trees where certain paths result in the same test being applied more than once
     - so need to detect those occurences and rewrite tree accordingly
       - i.e. find the bit where it's duplicated and promote that node to the top
       - could do with general function for promoting a node from anywhere within the tree?
   - also add support for children with differing tests
     - just need to behave as if they were leaf nodes?
     - brute force approach of promoting each test in turn until the best one is found
 - need to fix node/encoding dedupe code to increase/decrease reference counts by appropriate amount
   - tricky, since dedupe callers need to retain a reference to the thing they're deduping?
   - or can be handled by listening out for a 'is dead' response and setting a null pointer?
 - support for external attributes
   - i.e. constraints which test stuff which isn't part of the instruction, e.g. processor mode
   - would need encoding files to give a list of the attributes and their possible values
     - and any constraints between the attributes?
   - then would need to upgrade opcode_subset to contain a list of the values of those attributes, as they are decoded?
*/

uint32_t opcodesize = 0;
uint32_t opcodebits;

FILE *fout = NULL;
FILE *fgraph = NULL;
FILE *fcache = NULL;

encodingset allencodings;
actionset allactions;
bool verbose = false;
uint32_t maxmaskbits = 5;
bool fuzzy = false;
bool incomplete = false;
bool valid = false;
const char *prefix="decgen__";
bool isvoid = false;
uint32_t maxpreextract = 0;
bool treewalker = false;
bool useargstruct = false;
map<string,argument> argstruct;
bool readcache;
bool writecache;
bool updatecache;
bool badcache;
constraintset cacheconstraints;


/*

				FRONTEND

*/

bool loadencodings(const char *file,const macroset &macros)
{
	if(!input_open(file))
	{
		fprintf(stderr,"Failed to open encodings file: %s\n",file);
		return false;
	}
	if(verbose) printf("Loading encodings from file: %s\n",file);
	while(!input_eof())
	{
		encoding *e = encoding::parse(macros);
		if(e)
			allencodings.push_back(e);
	}
	input_close();
	return true;
}

bool checkambiguitycompleteness(const encodingset &e,const opcode_subset *o,int level)
{
	if(verbose && (level < 5) && (o->mask != opcodebits))
	{
		printf("\r%x/%x",o->fixed,opcodebits);
		fflush(stdout);
	}
	if(!e.size())
	{
		if(incomplete)
			return true;
		fprintf(stderr,"\rUnable to decode subset with fixed %08x mask %08x\n",o->fixed,o->mask);
		return false;
	}
	else if(e.size()==1)
	{
		if(incomplete)
			return true;
		uint32_t opcode=0;
		uint32_t fixed = o->fixed;
		uint32_t mask = o->mask | (opcodebits & ~e[0]->getfields()); /* Mark the non-constrained fields as fixed - they won't effect our evaluation */
		do {
			if(!e[0]->matches(opcode | fixed))
			{
				fprintf(stderr,"\rUnable to decode: %08x\n",opcode | fixed);
				return false;
			}
			opcode++;
			while(opcode & mask)
				opcode = (opcode+mask)&~mask;
		} while(opcode & opcodebits);
		return true;
	}
	uint32_t bitcount=4;
	uint32_t startbit=opcodesize-bitcount*level;
	if(startbit >= opcodesize)
	{
		bitcount += startbit;
		startbit = 0;
	}
	if(!bitcount)
	{
		/* We've run out of bits, and yet have multiple encodings. Ambiguity! */
		fprintf(stderr,"\rAmbiguous encodings:");
		for(size_t i=0;i<e.size();i++)
		{
			string d = e[i]->describe();
			fprintf(stderr,"%s%s'",(i?" and '":" '"),d.c_str());
		}
		fprintf(stderr," for opcode %08x\n",o->fixed);
		return false;
	}
	encodingset *e2 = split_on_field_fast(o,e,startbit,bitcount);
	uint32_t size = ((uint32_t)1)<<bitcount;
	for(uint32_t i=0;i<size;i++)
	{
		opcode_subset o2 = *o;
		o2.constrain(i<<startbit,(size-1)<<startbit);
		if(!checkambiguitycompleteness(e2[i],&o2,level+1))
		{
			delete[] e2;
			return false;
		}
	}
	delete[] e2;
	return true;
}

bool loadactions(const char *file)
{
	if(!input_open(file))
	{
		fprintf(stderr,"Failed to open actions file: %s\n",file);
		return false;
	}
	if(verbose) printf("Loading actions from file: %s\n",file);
	while(!input_eof())
	{
		actiondef *a = actiondef::parse();
		if(a)
			allactions.push_back(a);
	}
	input_close();
	return true;
}

bool checkactions(actionset &a)
{
	for(size_t i=0;i<a.size();i++)
		for(size_t j=i+1;j<a.size();j++)
		{
			if(a[i]->hasname(a[j]->names[0]))
			{
				fprintf(stderr,"Action %s is multiply defined\n",a[j]->names[0]);
				return false;
			}
			if(a[i]->trymerge(a[j]))
			{
				a.erase(a.begin()+j);
				j--;
			}
		}
	return true;
}

bool resolveactions(encodingset &e,const actionset &a,const actiondef *def)
{
	for(size_t i=0;i<e.size();i++)
	{
		if(e[i]->action)
			continue;
		for(size_t j=0;j<a.size();j++)
			if(a[j]->hasname(e[i]->name))
			{
				if(!a[j]->tryassign(e[i]))
				{
					string c = e[i]->describe();
					fprintf(stderr,"Action %s incompatible with encoding '%s'\n",e[i]->name,c.c_str());
					return false;
				}
				break;
			}
		if(!e[i]->action)
		{
			if(def)
			{
				if(!def->tryassign(e[i]))
				{
					fprintf(stderr,"Default action incompatible with encoding %s\n",e[i]->name);
					return false;
				}
			}
			else
			{
				string c = e[i]->describe();
				fprintf(stderr,"No action found for encoding '%s'\n",c.c_str());
				return false;
			}
		}
	}
	return true;
}

void resolvekin(encodingset &e)
{
	for(size_t i=0;i<e.size();i++)
	{
		/* e[i] will become the root of any kin discovered, so skip it if it's already a child */
		if(!e[i]->kin)
		{
			for(size_t j=i+1;j<e.size();j++)
				if(!e[j]->kin && e[i]->samecode(e[j])) /* If e[j] has kin then it can't possibly be kin with e[i] */
				{
					e[j]->kin = e[i];
					e[j]->kin_next = e[i]->kin_next;
					e[i]->kin_next = e[j];
				}
		}
	}
}

node *build_tree(const opcode_subset *o,const encodingset &e)
{
	if(e.size() == 0)
	{
		/* dead-end paths can and do happen. Try not to worry about it! */
		return NULL;
	}
	node *n = determine_best_node(o,e);
	if(!n)
		return NULL;
	if(n->gettype() == ELEAF)
		return n;
	node **children = new node *[n->getnumchildren()];
	for(uint32_t i=0;i<n->getnumchildren();i++)
	{
		children[i] = build_tree(n->getsubset(i),n->getchildset(i));
	}
	n->setchildren(children);
	/* If we just built a mask node, deduplicate it to save on memory */
	if(n->gettype() == EMASK)
		dedupe_tree(n,false);
	return n;
}

static void declare_argstruct()
{
	fprintf(fout,"typedef struct {\n");
	for(map<string,argument>::const_iterator it=argstruct.begin();it!=argstruct.end();++it)
	{
		const argument &arg = it->second;
		switch(arg.type)
		{
		case EARG_FIELD:
			{
				uint32_t count=0;
				for(size_t i=0;i<arg.fields.size();i++)
					count += arg.fields[i]->bitcount;
				if(count <= 8)
					fprintf(fout,"\tuint8_t %s;\n",it->first.c_str());
				else if(count <= 16)
					fprintf(fout,"\tuint16_t %s;\n",it->first.c_str());
				else
					fprintf(fout,"\tuint32_t %s;\n",it->first.c_str());
			}
			break;
		case EARG_NONSTANDARD:
			fprintf(fout,"\tuint32_t %s;\n",it->first.c_str());
			break;
		default:
			assert(0);
		}
	}
	fprintf(fout,"} %sargstruct;\n",prefix);
}

static void emit_argstruct(const char *opcodevar)
{
	fprintf(fout,"\t%sargstruct *%sargs,%stheargs;\n",prefix,prefix,prefix);
	for(map<string,argument>::const_iterator it=argstruct.begin();it!=argstruct.end();++it)
	{
		const argument &arg = it->second;
		fprintf(fout,"\t%stheargs.%s",prefix,it->first.c_str());
		switch(arg.type)
		{
		case EARG_FIELD:
			{
				uint32_t offset = 0;
				for(size_t i=0;i<arg.fields.size();i++)
				{
					fprintf(fout," %c (((%s >> %d) & 0x%x)<<%d)",(i?'|':'='),opcodevar,arg.fields[i]->startbit,((((uint32_t)1)<<arg.fields[i]->bitcount)-1),offset);
					offset += arg.fields[i]->bitcount;
				}
				fprintf(fout,";\n");
			}
			break;
		case EARG_NONSTANDARD:
			fprintf(fout," = ((%s & 0x%x) ^ 0x%x);\n",opcodevar,arg.shouldmatch_mask,arg.shouldmatch_value);
			break;
		default:
			assert(0);
		}
	}
	fprintf(fout,"\t%sargs = &%stheargs;\n",prefix,prefix);
}

static void write_bits(uint32_t count,uint32_t value,int &acc,int &acc_shift)
{
	while(count--)
	{
		acc |= (value & 1)<<acc_shift;
		acc_shift++;
		if(acc_shift == 8)
		{
			fprintf(fout,"\t0x%02x,\n",acc);
			acc = acc_shift = 0;
		}
		value >>= 1;
	}
}

static uint32_t numbitsfor(uint32_t value)
{
	uint32_t ret=0;
	while(value)
	{
		value >>= 1;
		ret++;
	}
	return ret;
}

bool save_tree(node *n,const char *file,bool append,const char *pre,const char *post,const char *name)
{
	fout = fopen(file,(append?"a":"w"));
	if(!fout)
	{
		fprintf(stderr,"Couldn't open output file: %s\n",file);
		return false;
	}
	if(pre)
	{
		FILE *p = fopen(pre,"r");
		if(!p)
		{
			fprintf(stderr,"Couldn't open pre file: %s\n",pre);
			return false;
		}
		char buf[4096];
		while(!feof(p))
		{
			size_t count = fread(buf,1,sizeof(buf),p);
			if(ferror(p))
			{
				fprintf(stderr,"Error reading from pre file: %s\n",pre);
				return false;
			}
			fwrite(buf,count,1,fout);
		}
		fclose(p);
	}
	if(useargstruct)
		declare_argstruct();
	if(treewalker)
	{
		int acc=0,acc_shift=0;

		int numleaf=0,numconstraint=0,nummask=0;
		vector<node *> leafnodes,constraintnodes,masknodes;
		/* Number the leaves first */
		n->treewalk_number(ELEAF,&numleaf,leafnodes);
		/* Then constraints */
		numconstraint = numleaf;
		n->treewalk_number(ECONSTRAINT,&numconstraint,constraintnodes);
		/* Then masks */
		nummask = numconstraint;
		n->treewalk_number(EMASK,&nummask,masknodes);
		/* Emit prototypes & bodies for leaf nodes, followed by the lookup table */
		for(size_t i=0;i<leafnodes.size();i++)
		{
			leafnodes[i]->emit_prototype();
			leafnodes[i]->emit_body();
		}
		fprintf(fout,"typedef RETURNTYPE (*%sleaffunction)(PARAMDECL",prefix);
		if(useargstruct)
			fprintf(fout,", const %sargstruct *%sargs",prefix,prefix);
		fprintf(fout,");\n");
		fprintf(fout,"static const %sleaffunction %sleaves[] = {\n",prefix,prefix);
		for(size_t i=0;i<leafnodes.size();i++)
		{
			string name = leafnodes[i]->get_name();
			fprintf(fout,"\t%s%s,\n",prefix,name.c_str());
		}
		fprintf(fout,"};\n");

		char opcodevar[strlen(prefix)+7];
		sprintf(opcodevar,"%sopcode",prefix);
		/* Number the field constraints */
		int numfield=0,numconst=0,numfunc=0;
		vector<constraint *> fields,consts,funcs;
		for(size_t i=0;i<constraintnodes.size();i++)
		{
			node_constraint *node = (node_constraint *) constraintnodes[i];
			constraint *c = (constraint *) node->getconstraint();
			c->treewalk_number(EFIELD,&numfield,fields);
		}
		/* Number the const constraints */
		numconst = numfield;
		for(size_t i=0;i<constraintnodes.size();i++)
		{
			node_constraint *node = (node_constraint *) constraintnodes[i];
			constraint *c = (constraint *) node->getconstraint();
			c->treewalk_number(ECONST,&numconst,consts);
		}
		/* Number the function constraints */
		numfunc = numconst;
		for(size_t i=0;i<constraintnodes.size();i++)
		{
			node_constraint *node = (node_constraint *) constraintnodes[i];
			constraint *c = (constraint *) node->getconstraint();
			c->treewalk_number(EFUNC,&numfunc,funcs);
		}
		int numconstraints = numfunc;
		/* Emit the field constraints */
		fprintf(fout,"static const uint8_t %sfields[] = {\n",prefix);
		for(size_t i=0;i<fields.size();i++)
		{
			constraint_field *c = (constraint_field *) fields[i];
			write_bits(numbitsfor(opcodesize-1),c->getstartbit(),acc,acc_shift);
			write_bits(numbitsfor(opcodesize-1),c->getbitcount()-1,acc,acc_shift);
		}
		if(acc_shift)
			fprintf(fout,"\t0x%02x\n",acc);
		acc = acc_shift = 0;
		fprintf(fout,"};\n");
		/* Emit the constant constraints */
		uint32_t maxconst = 0;
		for(size_t i=0;i<consts.size();i++)
			maxconst |= ((constraint_const *)consts[i])->getvalue();
		if(maxconst < 256)
			fprintf(fout,"static const uint8_t %sconsts[] = {\n",prefix);
		else if(maxconst < 65536)
			fprintf(fout,"static const uint16_t %sconsts[] = {\n",prefix);
		else
			fprintf(fout,"static const uint32_t %sconsts[] = {\n",prefix);
		for(size_t i=0;i<consts.size();i++)
		{
			constraint_const *c = (constraint_const *) consts[i];
			fprintf(fout,"\t0x%x,\n",c->getvalue());
		}
		fprintf(fout,"};\n");
		/* Emit the function constraints */
		int numfunctions=0;
		vector<funcdef *> functions;
		for(size_t i=0;i<funcs.size();i++)
		{
			constraint_func *c = (constraint_func *) funcs[i];
			funcdef *f = c->getfunc();
			if(f->treewalk_funcid == -1)
			{
				f->treewalk_funcid = numfunctions++;
				functions.push_back(f);
				f->emit();
			}
		}
		int constraintbits = numbitsfor(numconstraints-1);
		int functionidbits = numbitsfor(numfunctions-1);
		fprintf(fout,"static const uint8_t %sfuncs[] = {\n",prefix);
		for(size_t i=0;i<funcs.size();i++)
		{
			constraint_func *c = (constraint_func *) funcs[i];
			funcdef *f = c->getfunc();
			write_bits(functionidbits,f->treewalk_funcid,acc,acc_shift);
			/* For now treat all functions as having MAX_FUNC_PARAMS children */
			for(size_t j=0;j<MAX_FUNC_PARAMS;j++)
			{
				if(j<f->getnumparam())
					write_bits(constraintbits,c->getparams()[j]->treewalk_constraintid,acc,acc_shift);
				else
					write_bits(constraintbits,0,acc,acc_shift);
			}
		}
		if(acc_shift)
			fprintf(fout,"\t0x%02x\n",acc);
		acc = acc_shift = 0;
		fprintf(fout,"};\n");
		fprintf(fout,"static const uint8_t %sfuncargs[] = {\n",prefix);
		for(int i=0;i<numfunctions;i++)
		{
			funcdef *f = functions[i];
			write_bits(numbitsfor(MAX_FUNC_PARAMS),f->getnumparam(),acc,acc_shift);
		}
		if(acc_shift)
			fprintf(fout,"\t0x%02x\n",acc);
		acc = acc_shift = 0;
		fprintf(fout,"};\n");

		/* Utility function for reading bits from the arrays */
		fprintf(fout,"static uint32_t %sreadbits(const uint8_t *array,uint32_t offset,uint32_t count)\n",prefix);
		fprintf(fout,"\
{\n\
\tuint32_t result=0;\n\
\toffset += count;\n\
\twhile(count--) {\n\
\t\toffset--;\n\
\t\tresult <<= 1;\n\
\t\tresult |= ((array[offset>>3])>>(offset&7)) & 1;\n\
\t}\n\
\treturn result;\n\
}\n");

		/* Emit the constraint evaluator */
		fprintf(fout,"static uint32_t %sevalconstraint(const uint32_t opcode,const uint32_t id)\n",prefix);
		fprintf(fout,"{\n");
		fprintf(fout,"\tif(id < %d) {\n",numfield);
		fprintf(fout,"\t\t/* Field constraint */\n");
		fprintf(fout,"\t\tuint32_t bitoffset = id*%d;\n",numbitsfor(opcodesize-1)*2);
		fprintf(fout,"\t\tconst uint32_t startbit = %sreadbits(%sfields,bitoffset,%d); bitoffset += %d;\n",prefix,prefix,numbitsfor(opcodesize-1),numbitsfor(opcodesize-1));
		fprintf(fout,"\t\tconst uint32_t bitcount = %sreadbits(%sfields,bitoffset,%d)+1;\n",prefix,prefix,numbitsfor(opcodesize-1));
		fprintf(fout,"\t\treturn (opcode>>startbit) & ((((uint32_t)1)<<bitcount)-1);\n");
		fprintf(fout,"\t} else if(id < %d) {\n",numconst);
		fprintf(fout,"\t\t/* Const constraint */\n");
		fprintf(fout,"\t\treturn %sconsts[id-%d];\n",prefix,numfield);
		fprintf(fout,"\t} else {\n");
		fprintf(fout,"\t\t/* Function constraint */\n");
		fprintf(fout,"\t\tuint32_t bitoffset = (id-%d)*%d;\n",numconst,functionidbits+constraintbits*MAX_FUNC_PARAMS);
		fprintf(fout,"\t\tconst uint32_t function = %sreadbits(%sfuncs,bitoffset,%d); bitoffset += %d;\n",prefix,prefix,functionidbits,functionidbits);
		fprintf(fout,"\t\tuint32_t args[%d];\n",MAX_FUNC_PARAMS);
		fprintf(fout,"\t\tconst uint32_t numargs = %sreadbits(%sfuncargs,function*%d,%d);\n",prefix,prefix,numbitsfor(MAX_FUNC_PARAMS),numbitsfor(MAX_FUNC_PARAMS));
		fprintf(fout,"\t\tfor(uint32_t i=0;i<numargs;i++) {\n");
		fprintf(fout,"\t\t\targs[i] = %sevalconstraint(opcode,%sreadbits(%sfuncs,bitoffset,%d));\n",prefix,prefix,prefix,constraintbits);
		fprintf(fout,"\t\t\tbitoffset += %d;\n",constraintbits);
		fprintf(fout,"\t\t}\n");
		fprintf(fout,"\t\tswitch(function) {\n");
		fprintf(fout,"\t\tdefault:\n");
		for(size_t i=0;i<functions.size();i++)
		{
			funcdef *f = functions[i];
			fprintf(fout,"\t\tcase %d:\n",(int)i);
			fprintf(fout,"\t\t\treturn %sfunc_%s(",prefix,f->getname());
			for(uint32_t j=0;j<f->getnumparam();j++)
			{
				if(j)
					putc(',',fout);
				fprintf(fout,"args[%d]",j);
			}
			fprintf(fout,");\n");
		}
		fprintf(fout,"\t\t}\n");
		fprintf(fout,"\t}\n");
		fprintf(fout,"}\n");

		/* Work out how much space is needed to store the mask nodes */
		int nummaskchildren = 0;
		for(size_t i=0;i<masknodes.size();i++)
		{
			nummaskchildren += masknodes[i]->getnumchildren();
		}
		int maskinfobits = numbitsfor(opcodesize-1)+numbitsfor(maxmaskbits-1);
		int nodeidbits = 0;
		int maskidstride;
		int maxnodeid;
		do {
			nodeidbits++;
			maskidstride = 1; /* What we divide the mask bit addresses by */
			if(!(nodeidbits % maskinfobits))
				maskidstride = maskinfobits;
			maxnodeid = numconstraint + (nummaskchildren*nodeidbits + masknodes.size()*maskinfobits)/maskidstride;
		} while(maxnodeid > (1<<nodeidbits));
		/* Renumber mask nodes to use their final IDs */
		int idoffset = 0;
		for(size_t i=0;i<masknodes.size();i++)
		{
			masknodes[i]->treewalk_nodeid = numconstraint + idoffset;
			idoffset += (maskinfobits + masknodes[i]->getnumchildren()*nodeidbits)/maskidstride;
		}
		/* Write out the list of constraint nodes */
		fprintf(fout,"static const uint8_t %sconstraintnodes[] = {\n",prefix);
		for(size_t i=0;i<constraintnodes.size();i++)
		{
			node_constraint *node = (node_constraint *) constraintnodes[i];
			write_bits(constraintbits,node->getconstraint()->treewalk_constraintid,acc,acc_shift);
			write_bits(nodeidbits,node->getchildnode(0)->treewalk_nodeid,acc,acc_shift);
			write_bits(nodeidbits,node->getchildnode(1)->treewalk_nodeid,acc,acc_shift);
		}
		if(acc_shift)
			fprintf(fout,"\t0x%02x\n",acc);
		acc = acc_shift = 0;
		fprintf(fout,"};\n");
		/* Write out the list of mask nodes */
		fprintf(fout,"static const uint8_t %smasknodes[] = {\n",prefix);
		for(size_t i=0;i<masknodes.size();i++)
		{
			node_mask *mask = (node_mask *) masknodes[i];
			write_bits(numbitsfor(opcodesize-1),mask->getstartbit(),acc,acc_shift);
			write_bits(numbitsfor(maxmaskbits-1),mask->getbitcount()-1,acc,acc_shift);
			for(uint32_t j=0;j<mask->getnumchildren();j++)
			{
				node *child = mask->getchildnode(j);
				write_bits(nodeidbits,(child?child->treewalk_nodeid:0),acc,acc_shift);
			}
		}
		if(acc_shift)
			fprintf(fout,"\t0x%02x\n",acc);
		acc = acc_shift = 0;
		fprintf(fout,"};\n");

		/* Now the tree walker code */
		fprintf(fout,"RETURNTYPE %s(PARAMDECL)\n{\n",name);
		fprintf(fout,"\tconst uint32_t %s = OPCODE;\n",opcodevar);
		if(useargstruct)
			emit_argstruct(opcodevar);
		fprintf(fout,"\tuint32_t %snodeid = %d;\n",prefix,n->treewalk_nodeid);
		fprintf(fout,"\twhile(%snodeid >= %d) {\n",prefix,numleaf);
		fprintf(fout,"\t\tif(%snodeid >= %d) {\n",prefix,numconstraint);
		fprintf(fout,"\t\t\t/* Mask node */\n");
		fprintf(fout,"\t\t\tuint32_t %sbitoffset = (%snodeid-%d)*%d;\n",prefix,prefix,numconstraint,maskidstride);
		fprintf(fout,"\t\t\tuint32_t %sstartbit = %sreadbits(%smasknodes,%sbitoffset,%d); %sbitoffset += %d;\n",prefix,prefix,prefix,prefix,numbitsfor(opcodesize-1),prefix,numbitsfor(opcodesize-1));
		fprintf(fout,"\t\t\tuint32_t %sbitcount = %sreadbits(%smasknodes,%sbitoffset,%d)+1; %sbitoffset += %d;\n",prefix,prefix,prefix,prefix,numbitsfor(maxmaskbits-1),prefix,numbitsfor(maxmaskbits-1));
		fprintf(fout,"\t\t\t%sbitoffset += %d*((%s>>%sstartbit) & ((((uint32_t)1)<<%sbitcount)-1));\n",prefix,nodeidbits,opcodevar,prefix,prefix);
		fprintf(fout,"\t\t\t%snodeid = %sreadbits(%smasknodes,%sbitoffset,%d);\n",prefix,prefix,prefix,prefix,nodeidbits);
		fprintf(fout,"\t\t} else {\n");
		fprintf(fout,"\t\t\t/* Constraint node */\n");
		fprintf(fout,"\t\t\tuint32_t %sbitoffset = (%snodeid-%d)*%d;\n",prefix,prefix,numleaf,constraintbits+2*nodeidbits);
		fprintf(fout,"\t\t\tuint32_t %sconstraint = %sreadbits(%sconstraintnodes,%sbitoffset,%d); %sbitoffset += %d;\n",prefix,prefix,prefix,prefix,constraintbits,prefix,constraintbits);
		fprintf(fout,"\t\t\tif(%sevalconstraint(%s,%sconstraint) == 0) %sbitoffset += %d;\n",prefix,opcodevar,prefix,prefix,nodeidbits);
		fprintf(fout,"\t\t\t%snodeid = %sreadbits(%sconstraintnodes,%sbitoffset,%d);\n",prefix,prefix,prefix,prefix,nodeidbits);
		fprintf(fout,"\t\t}\n");
		fprintf(fout,"\t}\n");
		fprintf(fout,"\t%s(%sleaves[%snodeid])(PARAMS",(isvoid?"":"return "),prefix,prefix);
		if(useargstruct)
			fprintf(fout," PARAMSCOMMA %sargs",prefix);
		fprintf(fout,");\n");
		fprintf(fout,"}\n");
	}
	else
	{
		n->emit_prototype();
		n->emit_body();
		fprintf(fout,"RETURNTYPE %s(PARAMDECL)\n{\n",name);
		if(useargstruct)
			emit_argstruct("OPCODE");
		n->emit_call();
		fprintf(fout,"}\n");
	}
	if(post)
	{
		FILE *p = fopen(post,"r");
		if(!p)
		{
			fprintf(stderr,"Couldn't open post file: %s\n",post);
			return false;
		}
		char buf[4096];
		while(!feof(p))
		{
			size_t count = fread(buf,1,sizeof(buf),p);
			if(ferror(p))
			{
				fprintf(stderr,"Error reading from post file: %s\n",post);
				return false;
			}
			fwrite(buf,count,1,fout);
		}
		fclose(p);
	}
	fclose(fout);
	return true;
}

bool save_graph(node *n,const char *file)
{
	fgraph = fopen(file,"w");
	if(!fgraph)
	{
		fprintf(stderr,"Couldn't open graph file: %s\n",file);
		return false;
	}
#ifdef YED
	fprintf(fgraph,"\
<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>\n\
<graphml xmlns=\"http://graphml.graphdrawing.org/xmlns\"\n\
    xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n\
    xmlns:y=\"http://www.yworks.com/xml/graphml\"\n\
    xsi:schemaLocation=\"http://graphml.graphdrawing.org/xmlns\n\
     http://www.yworks.com/xml/schema/graphml/1.1/ygraphml.xsd\">\n\
  <key for=\"node\" id=\"a\" yfiles.type=\"nodegraphics\"/>\n\
  <key for=\"edge\" id=\"b\" yfiles.type=\"edgegraphics\"/>\n\
<graph>\n");
#else
	fprintf(fgraph,"\
<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n\
<graphml xmlns=\"http://graphml.graphdrawing.org/xmlns\"\n\
    xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n\
    xsi:schemalocation=\"http://graphml.graphdrawing.org/xmlns\n\
     http://graphml.graphdrawing.org/xmlns/1.0/graphml.xsd\">\n\
<graph>\n");
#endif
	n->emit_graph();
	fprintf(fgraph,"</graph>\n</graphml>\n");
	fclose(fgraph);
	return true;
}

void dedupe_constraints(const encodingset &e)
{
	vector<constraint **> c;
	for(size_t i=0;i<e.size();i++)
		for(uint32_t j=0;j<e[i]->numconstraints;j++)
		{
			e[i]->constraints[j]->collect(c);
			c.push_back(&e[i]->constraints[j]);
		}
	if(verbose) printf("%u constraints\n",(uint32_t)c.size());
	/* Increase reference count of all constraints, in order to avoid them deleting themselves while we iterate over the set
	   We also need a second array pointing straight to the constraints, because the act of deduplication will result in c's pointers being remapped, preventing us from correctly dereferencing them once we're done */
	constraint *cc[c.size()];
	for(size_t i=0;i<c.size();i++)
	{
		cc[i] = *(c[i]);
		cc[i]->reference();
	}
	/* Now search for duplicates */
	bool deduped[c.size()];
	memset(deduped,0,sizeof(bool)*c.size());
	uint32_t count=(uint32_t)c.size();
	for(size_t i=0;i<c.size();i++)
	{
		if(deduped[i])
			continue;
		for(size_t j=i+1;j<c.size();j++)
		{
			if(deduped[j])
				continue;
			if(constraint::deduplicate(c[i],c[j]))
			{
				deduped[i] = deduped[j] = true;
				count--;
			}
		}
	}
	if(verbose) printf("%u left after deduplication\n",count);
	/* Go through and derefence everything. This should safely clean up the deduped items */
	for(size_t i=0;i<c.size();i++)
		if(constraint::dereference(cc[i]))
			count++;
}

bool checkcontradictions(const encodingset &e)
{
	vector<constraint *> c;
	for(size_t i=0;i<e.size();i++)
		for(uint32_t j=0;j<e[i]->numconstraints;j++)
		{
			constraint *cc = e[i]->constraints[j];
			size_t k;
			for(k=0;k<c.size();k++)
				if(c[k] == cc)
					break;
			if(k == c.size())
			{
				c.push_back(cc);
				assert(cc->getfields()); /* Obviously constant constraints should have been eliminated at parse time */
				uint32_t opcode=0;
				bool good = false;
				uint32_t mask = opcodebits & ~cc->getfields();
				do {
					if(cc->eval(opcode))
					{
						good = true;
						break;
					}
					opcode++;
					while(opcode & mask)
						opcode = (opcode+mask)&~mask;
				} while(opcode & opcodebits);
				if(!good)
				{
					string str = cc->describe(e[i]->fields,e[i]->numfields);
					string str2 = e[i]->describe();
					fprintf(stderr,"Constraint '%s' of encoding '%s' is always false\n",str.c_str(),str2.c_str());
					return false;
				}
			}
		}
	return true;
}

void dedupe_tree(node *tree,bool v)
{
	vector<node **> n;
	tree->collect(n);
	if(v) printf("%u nodes\n",(uint32_t)n.size());
	/* Increase reference count of all nodes, in order to avoid them deleting themselves while we iterate over the set
	   We also need a second array pointing straight to the nodes, because the act of deduplication will result in n's pointers being remapped, preventing us from correctly dereferencing them once we're done */
	node *nn[n.size()];
	for(size_t i=0;i<n.size();i++)
	{
		nn[i] = *(n[i]);
		nn[i]->reference();
	}
	/* Now search for duplicates */
	bool deduped[n.size()];
	memset(deduped,0,sizeof(bool)*n.size());
	uint32_t count=(uint32_t)n.size();
	for(size_t i=0;i<n.size();i++)
	{
		if(deduped[i])
			continue;
		for(size_t j=i+1;j<n.size();j++)
		{
			if(deduped[j])
				continue;
			if(node::deduplicate(n[i],n[j]))
			{
				deduped[i] = deduped[j] = true;
				count--;
			}
		}
	}
	if(v) printf("%u left after deduplication\n",count);
	/* Go through and derefence everything. This should safely clean up the deduped items */
	for(size_t i=0;i<n.size();i++)
		if(node::dereference(nn[i]))
			count++;
}

static void do_opt_strings(node *tree,bool v)
{
	/* Build a list of just the leaf nodes */
	vector<node_leaf *> n;
	{
		vector<node **> n2;
		tree->collect(n2);
		n.reserve(n2.size());
		for(size_t i=0;i<n2.size();i++)
		{
			node *n3 = *(n2[i]);
			if(n3->gettype() != ELEAF)
				continue;
			/* Have we already included this node? */
			size_t j;
			for(j=0;j<n.size();j++)
				if(n[j] == n3)
					break;
			if(j == n.size())
			{
				n.push_back((node_leaf *) n3);
			}
		}
	}
	if(v) printf("Performing opt_strings optimisation\nTree has %u leaf nodes\n",(uint32_t)n.size());
	/* Tokenise each action */
	vector<tokenset> tokensets;
	tokensets.reserve(n.size());
	for(size_t i=0;i<n.size();i++)
	{
		tokensets.push_back(tokenise(n[i]->get_encoding()->action->code));
	}
	/* Find compatible ones and merge them together */
	vector<tokenset> tomerge;
	vector<node_leaf *> nodes_tomerge;
	int final = (int) n.size();
	while(n.size())
	{
		tomerge.clear();
		nodes_tomerge.clear();
		tomerge.push_back(tokensets.back());
		nodes_tomerge.push_back(n.back());
		tokensets.pop_back();
		n.pop_back();
		for(size_t i=0;i<n.size();i++)
		{
			if(nodes_tomerge[0]->get_encoding()->samefields(n[i]->get_encoding()) /* Make sure both actions require the same fields from the opcode */
			 && compatible_tokensets(tomerge[0],tokensets[i])) /* Make sure action code is similar enough */
			{
				tomerge.push_back(tokensets[i]);
				nodes_tomerge.push_back(n[i]);
				tokensets.erase(tokensets.begin()+i);
				n.erase(n.begin()+i);
				i--;
			}
		}
		if(tomerge.size() > 1)
		{
			/* Create the optimised action object */
			opt_strings_action *opt_action = new opt_strings_action(nodes_tomerge[0]->get_encoding());
			/* Optimise the actions */
			vector<map<string,string> > arguments;
			merge_tokensets(tomerge,opt_action->code,arguments);
			/* Update the nodes to use this new action */
			for(size_t i=0;i<nodes_tomerge.size();i++)
			{
				nodes_tomerge[i]->set_opt_strings_action(opt_action,arguments[i]);
			}
			/* Update our count */
			final -= tomerge.size()-1;
		}
	}
	if(v) printf("Reduced down to %d leaf functions\n",final);
}

typedef struct {
	argument a;
	uint32_t count;
} counted_arg;

bool recurse_preextract(node *n,bool v)
{
	/* Collect unique child node pointers & process children */
	if(!n->getnumchildren())
		return false;
	bool didsomething = false;
	node *children[n->getnumchildren()];
	uint32_t numchildren=0;
	for(uint32_t i=0;i<n->getnumchildren();i++)
	{
		node *child = n->getchildnode(i);
		if(!child)
			continue;
		uint32_t j;
		for(j=0;j<numchildren;j++)
			if(children[j] == child)
				break;
		if(j == numchildren)
		{
			children[numchildren++] = child;
			didsomething |= recurse_preextract(child,v);
			/* Detect any args which we need to supply */
			for(map<string,argument>::iterator it=child->function_args.begin();it!=child->function_args.end();++it)
			{
				if(!n->local_vars.count(it->first) && !n->function_args.count(it->first))
				{
					if(v) printf("Force-promoting %s to %s\n",it->first.c_str(),n->get_name().c_str());
					n->local_vars[it->first] = it->second;
					didsomething = true;
				}
			}
		}
	}
	if((n->local_vars.size() >= maxpreextract) || (numchildren < 2))
		return didsomething;
	/* Get list of local vars, sort by use count */
	map<string,counted_arg> vars;
	size_t maxchildargs=0;
	for(uint32_t i=0;i<numchildren;i++)
	{
		if(children[i]->function_args.size() >= maxpreextract)
			return didsomething; /* Already too many args extracted from this child */
		maxchildargs = max(maxchildargs,children[i]->function_args.size());
		for(map<string,argument>::iterator it=children[i]->local_vars.begin();it!=children[i]->local_vars.end();++it)
		{
			/* Ignore null nonstandard fields */
			if((it->second.type == EARG_NONSTANDARD) && !it->second.shouldmatch_mask)
				continue;
			map<string,counted_arg>::iterator it2 = vars.find(it->first);
			if(it2 == vars.end())
			{
				counted_arg a;
				a.a = it->second;
				a.count = 1;
				vars[it->first] = a;
			}
			else
			{
				it2->second.count++;
			}
		}
	}
	if(!vars.size())
		return didsomething;
	/* TODO try and promote any?
	   For now only do if all require it */
	for(map<string,counted_arg>::iterator it=vars.begin();it!=vars.end();++it)
	{
		if(it->second.count == numchildren)
		{
			n->local_vars[it->first] = it->second.a;
			for(uint32_t i=0;i<numchildren;i++)
			{
				children[i]->function_args[it->first] = children[i]->local_vars[it->first]; /* Must transfer from local_vars to get the correct local variable name */
				children[i]->local_vars.erase(it->first);
			}
			didsomething = true;
			if(v) printf("Promoting %s to %s due to all %d children using it\n",it->first.c_str(),n->get_name().c_str(),numchildren);
			if(++maxchildargs >= maxpreextract)
				return didsomething;
		}
	}
	return didsomething;
}

void do_preextract(node *tree,bool v)
{
	/* Recurse through tree. if >1 child uses same local var, promote to parent. Repeat until we run out of changes to make. */
	while(recurse_preextract(tree,v)) {};
}

void prepare_argstruct(node *n)
{
	/* Add the args struct to this and all its children */
	char name[strlen(prefix)+5];
	sprintf(name,"%sargs",prefix);
	if(!n || n->function_args.count(name))
		return;
	argument a;
	a.type = EARG_ARGSTRUCT;
	a.str = name;
	a.shouldmatch_mask = a.shouldmatch_value = 0;
	n->function_args[name] = a;
	/* Add any local vars to the struct definition */
	for(map<string,argument>::iterator it=n->local_vars.begin();it!=n->local_vars.end();++it)
	{
		if(argstruct.count(it->first))
			continue;
		switch(it->second.type)
		{
		case EARG_NONSTANDARD:
			/* Ignore null nonstandard fields */
			if(!it->second.shouldmatch_mask)
				continue;
		case EARG_FIELD:
			/* Add it to the struct */
			argstruct[it->first] = it->second;
			break;
		default:
			assert(0);
		}
	}
	/* Process children */
	for(uint32_t i=0;i<n->getnumchildren();i++)
		prepare_argstruct(n->getchildnode(i));
}

void do_help()
{
	printf("\
decgen v1.41 2014/02/07                              Copyright (c) Jeffrey Lee\n\
Usage:\n\
decgen -bits=<n> [-D<name>=<val> [-D...]] -e <files...> [-valid] [-incomplete]\n\
       [-v] [-a <files...> [-o|-A] <outfile> [-pre <file>] [-post <file>]\n\
       [-prefix=<prefix>] [-default=<action>] [-name=<name>] [-verify] [-void]\n\
       [-maxmaskbits=<n>] [-fuzzy|-exact] [-graph <file>] [-opt-strings]\n\
       [-maxpreextract=<n>] [-treewalker] [-useargstruct]\n\
       [-readcache|-writecache|-updatecache <file>]]\n\
\n\
-bits=<n>          Specify number of bits in each opcode\n\
-D<name>=<val>     Define a macro for use in encoding files\n\
-e <files...>      Specify encoding files\n\
-valid             Disables the ambiguity and completeness checks; specify when\n\
                   processing known-good data in order to speed up decgen.\n\
-incomplete        Indicate that the encoding files describe an incomplete\n\
                   subset of the instruction set (disables completeness checks)\n\
-v                 Enable verbose output\n\
-a <files...>      Specify action files\n\
-o <outfile>       Specify output file\n\
-A <outfile>       Specify output file (append mode)\n\
-pre <file>        Specify file whose content is to be placed at the start of\n\
                   the output file\n\
-post <file>       Specify file whose content is to be placed at the end of the\n\
                   output file\n\
-prefix=<prefix>   Specify prefix used for all functions generated by decgen.\n\
                   Default is decgen__\n\
-default=<action>  Name of default action to invoke if none matches an encoding\n\
-name=<name>       Name of decoder function. Default is DECODER.\n\
-verify            Verify tree decodes correctly (for debugging generator; may\n\
                   be slow!)\n\
-maxmaskbits=<n>   Set maximum number of bits to consider when creating mask\n\
                   nodes (a.k.a. switch statements/jump tables). Default 5.\n\
-fuzzy             For incomplete instruction sets, indicates that the decoder\n\
                   is not required to have any predictable behaviour when asked\n\
                   to decode an opcode for which no definition exists.\n\
-exact             For incomplete instruction sets, indicates that the decoder\n\
                   should perform the specified -default action when it\n\
                   encounters an opcode for which no definition exists.\n\
                   Warning: Selecting this option will significantly increase\n\
                   the amount of time taken to generate a tree.\n\
-graph <file>      Name of file to write decision tree to in GraphML format\n\
-void              Indicate that RETURNTYPE has been #define'd to 'void', and\n\
                   therefore decgen shouldn't attempt to return the value of\n\
                   functions. Useful for supressing warnings/errors generated\n\
                   by fussy compilers.\n\
-opt-strings       Attempt to optimise generated code by merging together\n\
                   action functions if the only difference between two\n\
                   functions are the string literals that they use.\n\
-maxpreextract=<n> Specify the maximum number of fields to pre-extract from\n\
                   opcodes at higher levels of the tree. This is the maximum\n\
                   number that will be passed to any one child function; tune\n\
                   it to control code size/performance on the target platform.\n\
                   Default value is zero.\n\
-treewalker        Instead of emitting the tree as pure code, emit it in a\n\
                   compact binary form instead, along with a tree walker\n\
                   function.\n\
-useargstruct      Place all fields into a structure to avoid leaves having to\n\
                   extract them manually.\n\
-readcache <file>  Read cached tree from <file>.\n\
-writecache <file> Write cached tree to <file>.\n\
-updatecache <file> Use cache file and update it if corrupt/out-of-date.\n\
");
}

int main(int argc,char **argv)
{
	bool verify = false;
	const char *outfile = NULL,*def = NULL,*pre=NULL,*post=NULL,*name="DECODER",*graph=NULL,*cachefile=NULL;
	vector<const char*> encodings,actions,*insert=NULL;
	bool exact = false;
	macroset macros;
	bool append = false;
	bool opt_strings = false;
	/* Parse command line options */
	for(int i=1;i<argc;i++)
	{
		if(argv[i][0] == '-')
		{
			if(sscanf(argv[i],"-bits=%u",&opcodesize) == 1)
			{
				if(opcodesize==sizeof(opcodebits)*8)
					opcodebits = ~(uint32_t)0;
				else
					opcodebits = (((uint32_t)1)<<opcodesize)-1;
			}
			else if(!strcmp(argv[i],"-incomplete"))
				incomplete = true;
			else if(!strcmp(argv[i],"-help") || !strcmp(argv[i],"--help") || !strcmp(argv[i],"-h"))
			{
				do_help();
				return 0;
			}
			else if(!strcmp(argv[i],"-v"))
				verbose = true;
			else if(!strcmp(argv[i],"-e"))
			{
				if(i<argc-1)
					insert = &encodings;
			}
			else if(!strcmp(argv[i],"-a"))
			{
				if(i<argc-1)
					insert = &actions;
			}
			else if(!strcmp(argv[i],"-o"))
			{
				if(i<argc-1)
				{
					append = false;
					outfile = argv[++i];
				}
			}
			else if(!strcmp(argv[i],"-A"))
			{
				if(i<argc-1)
				{
					append = true;
					outfile = argv[++i];
				}
			}
			else if(!strcmp(argv[i],"-graph"))
			{
				if(i<argc-1)
					graph = argv[++i];
			}
			else if(!strncmp(argv[i],"-default=",9))
				def = argv[i] + 9;
			else if(!strcmp(argv[i],"-valid"))
				valid = true;
			else if(!strcmp(argv[i],"-verify"))
				verify = true;
			else if(!strcmp(argv[i],"-exact"))
				exact = true;
			else if(!strcmp(argv[i],"-fuzzy"))
				fuzzy = true;
			else if(!strcmp(argv[i],"-pre"))
			{
				if(i<argc-1)
					pre = argv[++i];
			}
			else if(!strcmp(argv[i],"-post"))
			{
				if(i<argc-1)
					post = argv[++i];
			}
			else if(!strncmp(argv[i],"-prefix=",8))
				prefix = argv[i] + 8;
			else if(!strncmp(argv[i],"-name=",6))
				name = argv[i] + 6;
			else if((argv[i][1] == 'D') && strchr(argv[i],'='))
			{
				macro m;
				m.name = argv[i]+2;
				m.value = strchr(argv[i],'=')+1;
				m.value[-1] = 0;
				m.numparam = 0;
				/* Look for any parameters */
				char *c = strchr(m.name,'(');
				char *params[31];
				uint32_t u;
				if(c)
				{
					do {
						*c = 0;
						if(m.numparam==31)
						{
							fprintf(stderr,"Macro '%s' takes too many arguments\n",m.name);
							return -1;
						}
						params[m.numparam++] = ++c;
						while(*c && (*c != ',') && (*c != ')'))
							c++;
						if(!*c)
						{
							fprintf(stderr,"Macro '%s' has unterminated argument list\n",m.name);
							return -1;
						}
						if(!c[-1])
						{
							fprintf(stderr,"Macro '%s' has argument with no name\n",m.name);
							return -1;
						}
						/* TODO - Check for duplicated argument names */
					} while(*c != ')');
					*c = 0;
				}
				m.namelen = strlen(m.name);
				/* Convert the value part to its tokenised form */
				c = m.value;
				while(*c)
				{
					if(*c == '\\')
						memmove(c,c+1,strlen(c+1)+1);
					else if(*c == '[')
					{
						char *c2 = strchr(c+1,']');
						if(!c2)
						{
							fprintf(stderr,"']' expected when parsing macro '%s' value\n",m.name);
							return -1;
						}
						*c2++ = 0;
						for(u=0;u<m.numparam;u++)
							if(!strcmp(params[u],c+1))
								break;
						if(u==m.numparam)
						{
							fprintf(stderr,"Macro '%s' does not accept a '%s' parameter\n",m.name,c+1);
							return -1;
						}
						*c = u+1;
						memmove(c+1,c2,strlen(c2)+1);
					}
					else if((*c > 0) && (*c < 32))
					{
						fprintf(stderr,"Illegal character in macro '%s'\n",m.name);
						return -1;
					}
					c++;
				}
				size_t j;
				for(j=0;j<macros.size();j++)
					if(!strcmp(macros[j].name,m.name) && (macros[j].numparam == m.numparam))
					{
						macros[j] = m;
						break;
					}
				if(j == macros.size())
					macros.push_back(m);
			}
			else if(!strcmp(argv[i],"-void"))
				isvoid = true;
			else if(!strcmp(argv[i],"-opt-strings"))
				opt_strings = true;
			else if(sscanf(argv[i],"-maxpreextract=%u",&maxpreextract) == 1)
			{
			}
			else if(!strcmp(argv[i],"-treewalker"))
				treewalker = true;
			else if(!strcmp(argv[i],"-useargstruct"))
				useargstruct = true;
			else if(!strcmp(argv[i],"-readcache"))
			{
				readcache = true;
				if(i<argc-1)
					cachefile = argv[++i];
			}
			else if(!strcmp(argv[i],"-writecache"))
			{
				writecache = true;
				if(i<argc-1)
					cachefile = argv[++i];
			}
			else if(!strcmp(argv[i],"-updatecache"))
			{
				updatecache = true;
				if(i<argc-1)
					cachefile = argv[++i];
			}
			else if(sscanf(argv[i],"-maxmaskbits=%u",&maxmaskbits) != 1)
			{
				fprintf(stderr,"Unrecognised option: %s\n",argv[i]);
				do_help();
				return 1;
			}
		}
		else if(insert)
			insert->push_back(argv[i]);
		else
		{
			fprintf(stderr,"Unrecognised option: %s\n",argv[i]);
			do_help();
			return 1;
		}
	}
	if(argc == 1)
	{
		do_help();
		return 1;
	}
	if((opcodesize == 0) || (opcodesize > sizeof(opcodebits)*8))
	{
		fprintf(stderr,"Bad opcode size: %u\n",opcodesize);
		do_help();
		return 1;
	}
	assert(opcodebits);
	if((!actions.size()) && (outfile || graph))
	{
		fprintf(stderr,"Output file specified but no actions file(s)\n");
		do_help();
		return 1;
	}
	if((actions.size()) && !(outfile || graph))
	{
		fprintf(stderr,"Actions specified but no output file\n");
		do_help();
		return 1;
	}
	if(!encodings.size())
	{
		fprintf(stderr,"No encodings file(s) specified\n");
		do_help();
		return 1;
	}
	if(!maxmaskbits || (maxmaskbits>opcodesize))
	{
		fprintf(stderr,"Bad max mask bits: %u\n",maxmaskbits);
		do_help();
		return 1;
	}
	if(useargstruct && maxpreextract)
	{
		fprintf(stderr,"-useargstruct can't be used with -maxpreextract\n");
		return 1;
	}
	if(incomplete && (outfile || graph))
	{
		if(exact && fuzzy)
		{
			fprintf(stderr,"Cannot produce a decoder that is both exact and fuzzy!\n");
			do_help();
			return 1;
		}
		else if(!exact && !fuzzy)
		{
			fprintf(stderr,"Must specify either -exact or -fuzzy when producing a decoder for incomplete instruction sets\n");
			do_help();
			return 1;
		}
		else if(exact && !def)
		{
			fprintf(stderr,"Default action must be specified when using -exact\n");
			do_help();
			return 1;
		}
	}
	else
		exact = fuzzy = false;
	if((readcache && writecache) || (readcache && updatecache) || (writecache && updatecache))
	{
		fprintf(stderr,"Only one cache option can be specified\n");
		do_help();
		return 1;
	}
	else if(readcache || writecache || updatecache)
	{
		if(!cachefile)
		{
			fprintf(stderr,"No cache filename specified\n");
			do_help();
			return 1;
		}
		if(updatecache)
			readcache = true;
		fcache = fopen(cachefile,(readcache?"rb":"wb"));
		if(!fcache && updatecache)
		{
			readcache = false;
			writecache = true;
			fcache = fopen(cachefile,"wb");
		}
		if(!fcache)
		{
			fprintf(stderr,"Failed to open cache file\n");
			return 1;
		}
	}
	/* Load encodings */
	for(size_t i=0;i<encodings.size();i++)
	{
		if(!loadencodings(encodings[i],macros))
			return 1;
	}
	encodings.clear();
	if(verbose) printf("%u encodings loaded\n",(uint32_t) allencodings.size());
	/* Dedupe constraints. This will allow the bulk of the code to just use simple pointer checks when determining if two constraints are identical. */
	if(verbose) printf("Deduplicating constraints...\n");
	dedupe_constraints(allencodings);
	/* Verify that there aren't any contradictory constraints (i.e. constraints which are always false) */
	if(!valid)
	{
		if(verbose) printf("Checking constraints for contradictions\n");
		if(!checkcontradictions(allencodings))
			return 1;
	}
	/* Check collisions & completeness */
	opcode_subset o;
	if(!valid)
	{
		if(verbose) printf("Checking ambiguity%s...\n",(incomplete?"":" and completeness"));
		if(!checkambiguitycompleteness(allencodings,&o,1))
			return 1;
		if(verbose) printf("\r                 \r");
	}
	/* Was that it? */
	if(!outfile && !graph)
	{
		if(verbose && !valid) printf("Encodings OK\n");
		return 0;
	}
	/* Load actions */
	for(size_t i=0;i<actions.size();i++)
	{
		if(!loadactions(actions[i]))
			return 1;
	}
	actions.clear();
	if(verbose) printf("%u actions loaded\n",(uint32_t) allactions.size());
	/* Check collisions and merge */
	if(!checkactions(allactions))
		return 1;
	if(verbose) printf("%u actions after merging\n",(uint32_t) allactions.size());
	/* Hunt for default action */
	actiondef *defaultaction=NULL;
	if(def)
	{
		for(size_t i=0;i<allactions.size();i++)
			if(allactions[i]->hasname(def))
			{
				defaultaction = allactions[i];
				break;
			}
		if(!defaultaction)
		{
			fprintf(stderr,"Unable to find default action '%s'\n",def);
			return 1;
		}
	}
	if(exact)
	{
		if(defaultaction->numparam)
		{
			fprintf(stderr,"Default action must be parameterless for exact-incomplete trees\n");
			return 1;
		}
		/* Compute a single encoding that represents the inverse of all the other encodings
		   This will likely have a significant performance effect on the tree generator, but it's the easiest way of ensuring the tree correctly classifies the undefined instructions */
		allencodings.push_back(encoding::generateinverse(allencodings,"(undefined)","(undefined)"));
		/* TODO - Adjust constraint dedupe code so we can call it again from here (currently asserts, and may do bad stuff 2nd time round in general) */
		if(verbose) printf("Deduplicating constraints...\n");
		dedupe_constraints(allencodings);
	}
	/* Assign to encodings */
	if(!resolveactions(allencodings,allactions,defaultaction))
		return 1;
	if(verbose) printf("Actions assigned OK\n");
	if(verbose) printf("Calculating kin...\n");
	resolvekin(allencodings);
	/* Now produce the decision tree */
	if(verbose) printf("Generating decision tree...\n");
	node *tree = build_tree(&o,allencodings);
	if(!badcache && readcache)
	{
		/* Check we're at EOF */
		char c;
		if(fread(&c,sizeof(c),1,fcache) == 1)
		{
			fprintf((updatecache?stdout:stderr),"Cache file corrupt or out of date\n");
			badcache = true;
		}
	}
	if(badcache && updatecache && readcache)
	{
		if(tree)
			node::dereference(tree);
		printf("Restarting tree generation with fresh cache file\n");
		fclose(fcache);
		fcache = fopen(cachefile,"wb");
		if(!fcache)
		{
			fprintf(stderr,"Failed to open cache file\n");
			return 1;
		}
		badcache = false;
		readcache = false;
		writecache = true;
		tree = build_tree(&o,allencodings);
	}
	if(!tree || badcache)
	{
		fprintf(stderr,"Tree generation failed\n");
		return 1;
	}
	if(verbose) printf("Tree generation complete\n");
	if(fcache)
	{
		fclose(fcache);
		fcache = NULL;
		readcache = writecache = false;
	}
	/* Dedupe tree */
	/* We only need to do this if the root node isn't a mask node, since mask nodes get automatically deduplicated after insertion */
	if(tree->gettype() != EMASK)
	{
		if(verbose) printf("Deduplicating tree nodes...\n");
		dedupe_tree(tree,verbose);
	}
	/* Shrink tree */
	if(verbose) printf("Shrinking tree...\n");
	node::shrink(tree);
	/* Deduplicate again just in case the shrinking missed something */
	dedupe_tree(tree,false);
	/* Prepare arguments */
	tree->prepare_args();
	/* Optional optimisations */
	if(opt_strings)
	{
		do_opt_strings(tree,verbose);
	}
	if(maxpreextract)
	{
		do_preextract(tree,verbose);
	}
	else if(useargstruct)
	{
		prepare_argstruct(tree);
	}
	if(verbose)
	{
		vector<node *> n;
		uint32_t min,max;
		tree->collect(n);
		tree->calcdepth(min,max);
		printf("Final tree contains %u nodes, min depth=%u max depth=%u\n",(uint32_t)n.size(),min,max);
	}
	if(verify)
	{
		if(verbose) printf("Verifying tree...\n");
		if(fuzzy)
		{
			/* Fuzzy trees must be verified on a per-encoding basis */
			for(size_t i=0;i<allencodings.size();i++)
			{
				const encoding *e3 = allencodings[i];
				if(verbose)
				{
					printf("\r%u/%u",(uint32_t)i,(uint32_t)allencodings.size());
					fflush(stdout);
				}
				uint32_t opcode = 0;
				do {
					if(e3->matches(opcode | e3->fixed))
					{
						const encoding *e = tree->verify(opcode | e3->fixed);
						if(!e)
						{
							fprintf(stderr,"\rTree verification failed for %x; tree failed to classify\n",opcode | e3->fixed);
							assert(0);
							return 1;
						}
						else
						{
							/* Check if e3 is e or any of its kin (tree optimisation means we might decode to a kin of the encoding and not the encoding itself) */
							if(e->kin) e = e->kin;
							if((e != e3) && (e != e3->kin))
							{
								string c = e->describe();
								fprintf(stderr,"\rTree verification failed for opcode %x; tree thought it matched root kin '%s'\n",opcode | e3->fixed,c.c_str());
								assert(0);
								return 1;
							}
						}
					}
					opcode++;
					while(opcode & e3->mask)
						opcode = (opcode+e3->mask)&~e3->mask;
				} while(opcode & opcodebits);
			}
		}
		else
		{
			/* Non-fuzzy trees can have the entire opcode set iterated over. Not sure if this is faster than the above method! */
			uint32_t opcode=0;
			const encoding *olde=NULL;
			do {
				if(verbose && !(opcode & (opcodebits>>8)))
				{
					printf("\r%x/%x",opcode,opcodebits);
					fflush(stdout);
				}
				const encoding *e = tree->verify(opcode);
				if(!e)
				{
					fprintf(stderr,"\rTree verification failed for %x; tree failed to classify\n",opcode);
					assert(0);
					return 1;
				}
				else
				{
					/* Check if we match e or any of its kin (tree optimisation means we might decode to a kin of the encoding and not the encoding itself) */
					if(e->kin) e = e->kin;
					if(!olde || ((olde != e) && (olde->kin != e) && !olde->matches(opcode))) /* Early-out if this is the same kin set as before, and the encoding that matched before matches this time */
					{
						const encoding *ee = e;
						olde = NULL;
						while(ee)
						{
							if(ee->matches(opcode))
							{
								olde = ee;
								break;
							}
							ee = ee->kin_next;
						}
						if(!ee)
						{
							string c = e->describe();
							fprintf(stderr,"\rTree verification failed for opcode %x; tree thought it matched root kin '%s'\n",opcode,c.c_str());
							assert(0);
							return 1;
						}
					}
				}
				opcode++;
			} while(opcode & opcodebits);
		}
		if(verbose) printf("\rTree good           \n");
	}
	if(outfile)
	{
		if(verbose) printf("Generating code\n");
		if(!save_tree(tree,outfile,append,pre,post,name))
			return 1;
	}
	if(graph)
	{
		if(verbose) printf("Writing graph\n");
		if(!save_graph(tree,graph))
			return 1;
	}
	node::dereference(tree);
	return 0;
}

bool cache_read(void *p,size_t s)
{
	if(fread(p,s,1,fcache) != 1)
	{
		fprintf((updatecache?stdout:stderr),"Failed to read from cache - out of date?\n");
		badcache = true;
		return false;
	}
	return true;
}

void cache_write(const void *p,size_t s)
{
	if(fwrite(p,s,1,fcache) != 1)
	{
		fprintf(stderr,"Failed to write to cache\n");
		exit(1);
	}
}
