/*
*Copyright(c)2014, Jeffrey Lee
*Allrightsreserved.
*
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
*/

#include "encoding.h"
#include "constraint.h"
#include "actiondef.h"
#include "funcdef.h"
#include "input.h"

encoding::encoding()
{
	name = text = NULL;
	fixed = mask = numfields = 0;
	shouldmatch_value = shouldmatch_mask = 0;
	fields = NULL;
	numconstraints = 0;
	constraints = NULL;
	action = NULL;
	kin = kin_next = NULL;
}

encoding::~encoding()
{
	delete name;
	delete text;
	if(fields)
	{
		for(uint32_t i=0;i<numfields;i++)
			delete fields[i].name;
		free(fields);
	}
	if(numconstraints)
	{
		for(uint32_t i=0;i<numconstraints;i++)
			constraint::dereference(constraints[i]);
		free(constraints);
	}
}

bool encoding::matches(uint32_t opcode) const
{
	if(fixed != (opcode & mask))
		return false;

	for(uint32_t i=0;i<numconstraints;i++)
		if(constraints[i]->eval(opcode) == 0)
			return false;
	return true;
}

uint32_t encoding::getfields() const
{
	uint32_t f = mask;
	for(uint32_t i=0;i<numconstraints;i++)
		f |= constraints[i]->getfields();
	return f;
}

void encoding::emit_comment() const
{
	string c;
	c = describe();
	fputs(c.c_str(),fout);
	if(!kin)
	{
		/* If this is a root kin (which currently it always will be), output the kin list */
		const encoding *k = kin_next;
		while(k)
		{
			c = k->describe();
			fprintf(fout,"\n     %s",c.c_str());
			k = k->kin_next;
		}
	}
}

bool encoding::samecode(const encoding *e) const
{
	/* Check actions match */
	if(action != e->action)
		return false;
	/* If the nonstandard bits are required, check their patterns match */
	if((action->nonstandard) && ((shouldmatch_value != e->shouldmatch_value) || (shouldmatch_mask != e->shouldmatch_mask)))
		return false;
	/* Check that all action-relevant fields are the same */
	for(uint32_t i=0;i<numfields;i++)
		if(fields[i].relevant)
		{
			for(uint32_t j=0;j<e->numfields;j++)
			{
				if(!strcmp(fields[i].name,e->fields[j].name))
				{
					if((fields[i].startbit != e->fields[j].startbit) || (fields[i].bitcount != e->fields[j].bitcount))
						return false;
					break;
				}
			}
		}
	return true;
}

bool encoding::samefields(const encoding *e) const
{
	/* Check that nonstandard masks are the same */
	if(action->nonstandard != e->action->nonstandard)
		return false;
	if((action->nonstandard) && ((shouldmatch_value != e->shouldmatch_value) || (shouldmatch_mask != e->shouldmatch_mask)))
		return false;
	/* Check that all action-relevant fields in this match fields in e */
	for(uint32_t i=0;i<numfields;i++)
		if(fields[i].relevant)
		{
			bool found = false;
			for(uint32_t j=0;j<e->numfields;j++)
			{
				if(!strcmp(fields[i].name,e->fields[j].name))
				{
					found = true;
					if((fields[i].startbit != e->fields[j].startbit) || (fields[i].bitcount != e->fields[j].bitcount) || (!e->fields[j].relevant))
						return false;
					break;
				}
			}
			if(!found)
				return false;
		}
	/* Check that there aren't any action-relevant fields in e that don't exist in this */
	for(uint32_t i=0;i<e->numfields;i++)
		if(e->fields[i].relevant)
		{
			bool found = false;
			for(uint32_t j=0;j<numfields;j++)
			{
				if(!strcmp(e->fields[i].name,fields[j].name))
				{
					found = true;
					break;
				}
			}
			if(!found)
				return false;
		}
	
	return true;
}

string encoding::describe() const
{
	string c(text);
	c.reserve(c.length()+1024); /* Keep reallocs to a minimum */
	/* Add constraints */
	for(uint32_t i=0;i<numconstraints;i++)
	{
		c += " {";
		c += constraints[i]->describe(fields,numfields);
		c += "}";
	}
	/* Add name */
	c += " ";
	c += name;
	/* And action if required */
	if(action && !action->hasname(name))
	{
		c += " (as ";
		c += action->names[0];
		c += ")";
	}
	return c;
}

void encoding::subtract(const encoding *e)
{
	/* e is an expression of the form:
	     ((ocpode & mask)==fixed) && c1 && c2 && c3 ...
	   we need to invert that:
	     !(((opcode & mask)==fixed) && c1 && c2 && c3...)
	   we can eliminate from the mask the bits which are fixed in both e and this, and any constraints in e that are also in this
	   but otherwise, there isn't much else we can do to simplify the constraint?
	   (except do more exhaustive checks to ensure that no element of the inversion is always true for this encoding) */
	uint32_t newfixed = e->mask &~mask; /* Bits which are fixed in e but not in this */
	constraint *additional[opcodesize + e->numconstraints]; /* Overestimate */
	uint32_t num=0;
	for(uint32_t startbit=0;startbit<opcodesize;startbit++)
	{
		uint32_t bitcount=0;
		while(newfixed & (((uint32_t)1)<<(startbit+bitcount)))
			bitcount++;
		if(bitcount)
		{
			constraint **p = new constraint *[2];
			p[0] = new constraint_field(startbit,bitcount);
			p[1] = new constraint_const((e->fixed>>startbit)&((((uint32_t)1)<<bitcount)-1));
			additional[num++] = new constraint_func(funcdef_eq,p);
			
			startbit += bitcount;
		}
	}
	for(uint32_t c=0;c<e->numconstraints;c++)
	{
		uint32_t d;
		for(d=0;d<numconstraints;d++)
			if(constraints[d]->equals(*e->constraints[c]))
				break;
		if(d == numconstraints)
		{
			e->constraints[c]->reference();
			additional[num++] = e->constraints[c];
		}
	}
	/* Now construct a constraint to tie all of 'additional' together */
	assert(num);
	while(num > 1)
	{
		constraint **p = new constraint *[2];
		p[0] = additional[--num];
		p[1] = additional[num-1];
		additional[num-1] = new constraint_func(funcdef_land,p);
	}
	constraint **p = new constraint *[1];
	p[0] = additional[0];
	constraints = (constraint **) realloc(constraints,sizeof(constraint *)*(numconstraints+1));
	assert(constraints);
	constraints[numconstraints++] = new constraint_func(funcdef_lnot,p);
}

encoding *encoding::parse(const macroset &macros)
{
	char buf[1024],*name,*temp;
	if (!input_getline(buf,sizeof(buf)))
		return NULL;
	/* Trim any comment to simplify latter checks & macro expansion */
	char *s = strchr(buf,'#');
	if(s)
		*s = 0;
	/* Macro expansion */
	s = buf;
	while((s=strchr(s,'['))!=0)
	{
		char *s2 = s;
		do {
			s2++;
		} while(*s2 && (*s2 != ']') && (*s2 != '('));

		if(!*s2)
		{
			input_errorc(s-buf,"Missing ']' when expanding macro\n");
			return NULL;
		}

		size_t len = s2-(s+1);
		char *params[31];
		uint32_t numparam=0;
		if(*s2 == '(')
		{
			/* Grab parameters */
			do {
				*s2 = 0;
				if(numparam==31)
				{
					input_errorc(s2-buf,"Too many macro parameters\n");
					return NULL;
				}
				params[numparam++] = ++s2;
				while(*s2 && (*s2 != ',') && (*s2 != ')'))
				{
					if((*s2 == '\\') && (s2[1]))
						s2++;
					s2++;
				}
				if(!*s2)
				{
					input_errorc(s-buf,"Missing ')' when expanding macro\n");
					return NULL;
				}
			} while (*s2 != ')');
			*s2++ = 0; /* Kill ')' */
		}
		*s2 = 0; /* Kill ']' */
		size_t i;
		for(i=0;i<macros.size();i++)
			if((macros[i].namelen == len) && !strcmp(macros[i].name,s+1) && (macros[i].numparam == numparam))
				break;
		if(i == macros.size())
		{
			input_errorc(s-buf,"Unknown macro '%s' with %u parameters\n",s+1,numparam);
			return NULL;
		}
		/* Check macro length */
		len = strlen(macros[i].value);
		char *m = macros[i].value;
		while(*m)
		{
			if((*m > 0) && (*m < 32))
				len += strlen(params[(*m)-1])-1;
			m++;
		}
		if((size_t)(s+len+strlen(s2+1)+1-buf) > sizeof(buf))
		{
			input_errorc(s-buf,"Line too long when expanding macro '%s'\n",s+1);
		}
		/* Create a temp buffer to expand into */
		char *o = temp = new char[len+1];
		m = macros[i].value;
		while(*m)
		{
			if((*m > 0) && (*m < 32))
			{
				char *p = params[(*m++)-1];
				while(*p)
				{
					if(*p == '\\')
						p++;
					*o++ = *p++;
				}
			}
			else
				*o++ = *m++;
		}
		*o = 0;
		assert((size_t)(o-temp) <= len);
		len = strlen(temp);
		/* Shuffle everything along */
		memmove(s+len,s2+1,strlen(s2+1)+1);
		/* Insert expansion */
		memcpy(s,temp,len);
		delete[] temp;
	}
	s = buf;
	while(*s && isspace(*s))
		s++;
	temp = s;
	if(*s == 0)
		return NULL;
	encoding *e = new encoding();
	int i = opcodesize-1;
	int len;
	while(*s && !isspace(*s))
	{
		switch(*s)
		{
		case '1':
			e->fixed |= ((uint32_t)1)<<i;
		case '0':
			e->mask |= ((uint32_t)1)<<i;
			i--;
			s++;
			break;
		case '(':
			s++;
			if((*s == '1') || (*s == '0'))
			{
				/* 'should match' field */
				while(*s != ')')
				{
					switch(*s)
					{
					case '1':
						e->shouldmatch_value |= ((uint32_t)1)<<i;
					case '0':
						e->shouldmatch_mask |= ((uint32_t)1)<<i;
						i--;
						s++;
						break;
					default:
						input_errorc(s-buf,"Unexpected character '%c' inside 'should match' field\n",*s);
						goto err;
					}
				}
				s++;
			}
			else if(*s == ':')
			{
				/* Anonymous field */
				s++;
				int count,size;
				if(sscanf(s,"%d%n",&count,&size) != 1)
				{
					input_errorc(s-buf,"Syntax error parsing anonymous field: Number expected\n");
					goto err;
				}
				i -= count;
				s += size;
				if(*s != ')')
				{
					input_errorc(s-buf,"Syntax error parsing anonymous field: ')' expected\n");
					goto err;
				}
				s++;
			}
			else if(isalpha(*s) || (*s == '_'))
			{
				/* Named field */
				int len=0;
				while(isalnum(s[len]) || (s[len] == '_'))
					len++;
				e->fields = (field *) realloc(e->fields,sizeof(field)*(e->numfields+1));
				assert(e->fields);
				field *f = &e->fields[e->numfields++];
				memset(f,0,sizeof(field));
				char *name = new char[len+1];
				assert(name);
				strncpy(name,s,len);
				name[len] = 0;
				f->name = name;
				s += len;
				if(*s == ')')
				{
					/* bool field */
					f->startbit = i--;
					f->bitcount = 1;
					s++;
				}
				else if(*s == ':')
				{
					int count,size;
					s++;
					if((sscanf(s,"%d%n",&count,&size) != 1) || (count < 1))
					{
						input_errorc(s-buf,"Syntax error parsing size of named field '%s': Number expected\n",f->name);
						goto err;
					}
					i -= count;
					f->startbit = i+1;
					f->bitcount = count;
					s += size;
					if(*s != ')')
					{
						input_errorc(s-buf,"Syntax error parsing size of named field '%s': ')' expected\n",f->name);
						goto err;
					}
					s++;
				}
				else
				{
					input_errorc(s-buf,"Unexpected character '%c' in named field\n",*s);
					goto err;
				}
			}
			else
			{
				input_errorc(s-buf,"Unexpected character '%c' inside field\n",*s);
				goto err;
			}
			break;
		default:
			if(isalpha(*s))
			{
				/* bool field */
				e->fields = (field *) realloc(e->fields,sizeof(field)*(e->numfields+1));
				assert(e->fields);
				field *f = &e->fields[e->numfields++];
				memset(f,0,sizeof(field));
				char *name = new char[2];
				assert(name);
				name[0] = *s;
				name[1] = 0;
				f->name = name;
				f->startbit = i;
				f->bitcount = 1;
				s++;
				i--;
			}
			else
			{
				input_errorc(s-buf,"Unexpected character '%c'\n",*s);
				goto err;
			}
			break;
		}
		/* Allow tabs to be used for formatting within the definition */
		if((i >= 0) && (*s == '\t'))
		{
			while(*s == '\t')
				s++;
			if(*s == '{')
				break; /* Will fall through to 'wrong length' case */
		}
	}
	if(i != -1)
	{
		input_error("Encoding is wrong length, should be %d bits but is actually %d bits\n",opcodesize,(opcodesize-1)-i);
		delete e;
		return NULL;
	}
	/* Verify field names are unique */
	for(i=0;(uint32_t)i<e->numfields;i++)
		for(uint32_t j=i+1;j<e->numfields;j++)
			if(!strcmp(e->fields[i].name,e->fields[j].name))
			{
				input_error("Duplicate field name: %s\n",e->fields[i].name);
				goto err;
			}
	/* Copy over definition string */
	name = new char[(s-temp)+1];
	assert(name);
	i=0;
	while(temp<s)
	{
		if(*temp != '\t') /* Gobble tabs */
			name[i++] = *temp;
		temp++;
	}
	name[i] = 0;
	e->text = name;
	/* Parse any constraints */
	while(*s && isspace(*s))
		s++;
	if(*s == 0)
	{
		input_error("Unexpected end-of-line when looking for constraints\n");
		goto err;
	}
	while(*s == '{')
	{
		s++;
		constraint *c = constraint::parse(e,s);
		if(!c)
		{
			input_errorc(s-buf,"Error parsing constraint\n");
			goto err;
		}
		if(!c->getfields())
		{
			/* Constraint value is constant, perhaps due to macro?
			   Eliminate it */
			if(!c->eval(0))
			{
				/* This encoding will never match anything, abort parsing now */
				constraint::dereference(c);
				delete e;
				return NULL;
			}
		}
		else
			simplify_and_insert(e,c);
		constraint::dereference(c);
		while(*s && isspace(*s))
			s++;
		if(*(s++) != '}')
		{
			input_errorc((s-1)-buf,"Syntax error: '}' expected\n");
			goto err;
		}

		while(*s && isspace(*s))
			s++;
		if(*s == 0)
		{
			input_error("Unexpected end-of-line when looking for constraints\n");
			goto err;
		}
	}
	/* Read encoding name */
	if(!isalnum(*s) && (*s != '_'))
	{
		input_errorc(s-buf,"Bad character '%c' when looking for encoding name\n",*s);
		goto err;
	}
	len=0;
	while(isalnum(s[len]) || (s[len] == '_'))
		len++;
	name = new char[len+1];
	assert(name);
	strncpy(name,s,len);
	name[len] = 0;
	s += len;
	e->name = name;
	while(*s && isspace(*s))
		s++;
	if((s[0] == 'i') && (s[1] == 'n') && isspace(s[2]))
	{
		/* This encoding is a subset of another encoding. Find out which! */
		s += 3;
		while(*s && isspace(*s))
			s++;
		if(!isalpha(*s) && (*s != '_'))
		{
			input_errorc(s-buf,"Bad character '%c' when processing 'in' specifier\n",*s);
			goto err;
		}
		temp = s;
		while(isalnum(*s) || (*s == '_'))
			s++;
		if(*s && !isspace(*s))
		{
			input_errorc(s-buf,"Bad character '%s' when processing 'in' specifier\n",*s);
			goto err;
		}
		/* Temporarily null terminate the string for ease of strcmp */
		len = *s;
		*s = 0;
		/* Now search for the indicated encoding */
		size_t j;
		for(j=0;j<allencodings.size();j++)
			if(!strcmp(temp,allencodings[j]->name))
				break;
		if(j == allencodings.size())
		{
			input_errorc(s-buf,"Couldn't find parent encoding '%s'\n",temp);
			delete e;
			return NULL;
		}
		*s = len;
		/* Now validate that e lies fully within j */
		encoding *parent = allencodings[j];
		if(!valid)
		{
			uint32_t opcode=0;
			uint32_t mask = e->mask | (opcodebits & (~e->getfields() & ~parent->getfields())); /* Skip the fixed bits in e, and the bits which are unconstrained in both */
			do {
				if(e->matches(opcode | e->fixed) && !parent->matches(opcode | e->fixed))
				{
					string a = e->describe();
					string b = parent->describe();
					input_error("Encoding '%s' falls outside parent '%s' on opcode %08x\n",a.c_str(),b.c_str(),opcode | e->fixed);
					delete e;
					return NULL;
				}
				opcode++;
				while(opcode & mask)
					opcode = (opcode+mask)&~mask;
			} while(opcode & opcodebits);
		}
		/* Now spot the constraints that differ, invert them, and add them to the parent */
		parent->subtract(e);
		while(*s && isspace(*s))
			s++;
	}
	if(*s == 0)
		return e;
	/* ... else something after the name, so it's an error */
err:
	delete e;
	return NULL;
}

encoding *encoding::generateinverse(const encodingset &e,const char *name,const char *text)
{
	encoding *n = new encoding();
	char *temp = new char[strlen(name)+1];
	assert(temp);
	strcpy(temp,name);
	n->name = name;
	temp = new char[strlen(text)+1];
	assert(temp);
	strcpy(temp,text);
	n->text = text;
	/* Start off matching against everything */
	n->fixed = 0;
	n->mask = 0;
	/* Now subtract each member of e in turn */
	for(size_t i=0;i<e.size();i++)
		n->subtract(e[i]);
	return n;
}
