/*
*Copyright(c)2014, Jeffrey Lee
*Allrightsreserved.
*
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
*/

#include "node.h"
#include "encoding.h"
#include "opcodesubset.h"
#include "actiondef.h"
#include "constraint.h"
#include "decgen.h"

static bool sort1(collecti a,collecti b)
{
	if(a.n != b.n)
		return (a.n < b.n);
	return (a.i < b.i);
}

static bool sort2(collecti a,collecti b)
{
	return (a.i < b.i);
}

opt_strings_action::opt_strings_action(const encoding *e)
{
	emitted = false;
	enc = e;
}

opt_strings_action::~opt_strings_action()
{
}

void opt_strings_action::emit(node_leaf *parent)
{
	if(emitted)
		return;
	emitted = true;
	/* Generate the argument list */
	for(size_t i=0;i<code.size();i++)
	{
		if(code[i].type == ESTRINGREF)
			args.push_back(code[i].value);
	}
	/* Emit prototype */
	fprintf(fout,"static RETURNTYPE %s%s(PARAMDECL",prefix,get_name().c_str());
	parent->emit_args();
	if(parent->function_args.size())
		fprintf(fout,",");
	else
		fprintf(fout," PARAMSCOMMA ");
	for(size_t i=0;i<args.size();i++)
	{
		fprintf(fout,"const char *%s%c",args[i].c_str(),(i==args.size()-1)?')':',');
	}
	fprintf(fout,"\n{\n");
	parent->emit_locals();
	/* Now emit body */
	for(size_t i=0;i<code.size();i++)
		fputs(code[i].value.c_str(),fout);
	fprintf(fout,"}\n");
}

string opt_strings_action::get_name() const
{
	/* Just use our address to generate a unique name */
	char buf[128];
	sprintf(buf,"optstrings_%p",this);
	return string(buf);
}

node::node(uint32_t num,const opcode_subset *o,const encodingset *res)
{
	refcount = 1;
	numchildren = num;
	results = res;
	children = NULL;
	if(num)
	{
		childset = new opcode_subset[num];
		/* Initialise child sets, but don't constrain */
		while(num--)
			childset[num] = *o;
	}
	else
		childset = NULL;
	emitted_prototype = emitted_body = emitted_graph = prepared_args = false;
	treewalk_nodeid = -1;
}

node::~node()
{
	assert(refcount == 0);
	delete[] results;
	if(children)
	{
		for(uint32_t i=0;i<numchildren;i++)
			if(children[i])
				dereference(children[i]);
		delete[] children;
	}
	delete[] childset;
}

uint32_t node::getnumchildren() const
{
	return numchildren;
}

const encodingset &node::getchildset(uint32_t i) const
{
	return results[i];
}

const opcode_subset *node::getsubset(uint32_t i) const
{
	return &childset[i];
}

const node *node::getchildnode(uint32_t i) const
{
	return children[i];
}

node *node::getchildnode(uint32_t i)
{
	return children[i];
}

void node::setchildren(node **arr)
{
	children = arr;
}

void node::collect(vector<node **> &n) const
{
	vector<collecti> c;
	/* Collect all our children, including duplicates */
	collect_int(c);
	/* Sort by node pointer, sub-sorted by original index */
	sort(c.begin(),c.end(),sort1);
	/* Iterate through and remove all the duplicate nodes
	   Iterate backwards to keep things fast
	   Except GCC's vector::erase doesn't seem to allow indices or reverse iterators as arguments
	   So build a new vector instead! */
	vector<collecti> c2;
	c2.reserve(c.size());
	node **last = NULL;
	for(size_t i=0;i<c.size();i++)
		if(c[i].n != last)
		{
			last = c[i].n;
			c2.push_back(c[i]);
		}
	c.clear();
	/* Now re-sort by index */
	sort(c2.begin(),c2.end(),sort2);
	/* Transfer to destination vector */
	n.reserve(c2.size());
	for(size_t i=0;i<c2.size();i++)
		n.push_back(c2[i].n);
}

void node::collect_int(vector<collecti> &c) const
{
	for(uint32_t i=0;i<numchildren;i++)
	{
		if(!children[i])
			continue;
		children[i]->collect_int(c);
		collecti ci;
		ci.n = &children[i];
		ci.i = c.size();
		if(c.capacity()==c.size())
			c.reserve(128);
		c.push_back(ci);
	}
}

void node::collect(vector<node *> &n) const
{
	/* Just use a simple algorithm, since this should only get used for small trees */
	for(uint32_t i=0;i<numchildren;i++)
	{
		if(!children[i])
			continue;
		uint32_t j;
		for(j=0;j<n.size();j++)
			if(n[j] == children[i])
				break;
		if(j == n.size())
		{
			children[i]->collect(n);
			if(n.capacity()==n.size())
				n.reserve(128);
			n.push_back(children[i]);
		}
	}
}

bool node::deduplicate(node **a,node **b)
{
	if((*a == *b) || !(**a == **b))
		return false;
	/* Merge into *a */
	dereference(*b);
	*b = *a;
	(*a)->refcount++;
	return true;
}

bool node::dereference(node *n)
{
	if(--(n->refcount))
		return false;
	delete n;
	return true;
}

void node::reference()
{
	refcount++;
}

void node::calcdepth(uint32_t &min,uint32_t &max)
{
	if(!numchildren)
	{
		min = max = 1;
		return;
	}
	min = ~0;
	max = 0;
	for(uint32_t i=0;i<numchildren;i++)
	{
		if(!children[i])
			continue;
		uint32_t a,b;
		children[i]->calcdepth(a,b);
		if(a<min)
			min = a;
		if(b>max)
			max = b;
	}
	min++;
	max++;
}

void node::shrink(node *&n)
{
	do {
		if(!n || (n->gettype() == ELEAF))
			return;
		uint32_t i;
		node *n2 = 0;
		for(i=0;i<n->numchildren;i++)
		{
			if(!n->children[i])
				continue;
			if(!n2)
				n2 = n->children[i];
			else if(n->children[i] != n2)
				break;
		}
		if(i != n->numchildren)
			break;
		/* all children are identical; node is redundant */
		if(n2) /* PROMOTE_CHILDREN means we might end up with no children? */
			n2->reference();
		dereference(n);
		n = n2;
		/* Loop back around and check n's children again */
	} while(1);
	/* Children differ; attempt to shrink them */
	/* If this is a mask node, shrink the mask area to minimum */
	if(n->gettype() == EMASK)
		((node_mask *) n)->shrink2();
	for(uint32_t i=0;i<n->numchildren;i++)
		if(n->children[i])
			shrink(n->children[i]);
#ifdef PROMOTE_CHILDREN
	/* All our children have been shrunk, now check if promoting our children's test will result in an overall reduction in tree size (i.e node count). Reducing the node count like this can help increase code performance, even if it adds an extra layer to some paths.
	   For the moment, only deal with the simplest of case, where all our children perform the same test (ignoring children which are leaves)
	   The easiest way of performing the test is to build a 2D array mapping the two tests to resulting nodes, then check how many unique nodes are needed for each combination */
	node *first=NULL;
	node ***array=NULL;
	for(uint32_t i=0;i<n->numchildren;i++)
		if(n->children[i] && (n->children[i]->numchildren))
		{
			first = n->children[i];
			array = new node **[first->numchildren];
			for(uint32_t j=0;j<first->numchildren;j++)
			{
				array[j] = new node *[n->numchildren];
				memset(array[j],0,sizeof(node*)*n->numchildren);
			}
			break;
		}
	if(!first)
		return; /* We have no non-leaf children, so there's nothing to promote */
	for(uint32_t i=0;i<n->numchildren;i++)
		if(n->children[i])
		{
			if(!n->children[i]->numchildren)
			{
				for(uint32_t j=0;j<first->numchildren;j++)
					array[j][i] = n->children[i];
			}
			else if(!first->sametest(*n->children[i]))
			{
				/* Tests differ - can't optimise this */
				for(uint32_t j=0;j<first->numchildren;j++)
					delete[] array[j];
				delete[] array;
				return;
			}
			else
			{
				for(uint32_t j=0;j<first->numchildren;j++)
					array[j][i] = n->children[i]->children[j];
			}
		}
	/* Build the new tree, then optimise it and compare the node count. This probably isn't the fastest way of doing everything!
	   This will also break the consitency of the 'subset' and 'childset' members, but luckily they aren't important anymore */
	opcode_subset o;
	node **newchildren = new node *[first->numchildren];
	for(uint32_t i=0;i<first->numchildren;i++)
	{
		for(uint32_t j=0;j<n->numchildren;j++)
		{
			if(array[i][j])
				array[i][j]->reference();
		}
		if(n->gettype()==EMASK)
			newchildren[i] = new node_mask(static_cast<node_mask *>(n)->getstartbit(),static_cast<node_mask *>(n)->getbitcount(),&o,NULL);
		else
			newchildren[i] = new node_constraint(static_cast<node_constraint *>(n)->getconstraint(),&o,NULL);
		/* Fill in the children */
		newchildren[i]->setchildren(array[i]);
	}
	delete[] array;
	node *newparent;
	if(first->gettype()==EMASK)
		newparent = new node_mask(static_cast<node_mask *>(first)->getstartbit(),static_cast<node_mask *>(first)->getbitcount(),&o,NULL);
	else
		newparent = new node_constraint(static_cast<node_constraint *>(first)->getconstraint(),&o,NULL);
	newparent->setchildren(newchildren);
	/* Optimise this new tree */
	dedupe_tree(newparent,false);
	/* Shrink the new children (but not the parent, otherwise we'll just sit in an infinite loop!) */
	for(uint32_t i=0;i<newparent->numchildren;i++)
		node::shrink(newparent->children[i]);
	/* Count unique nodes in each */
	vector<node *> n2,n3;
	n->collect(n2);
	newparent->collect(n3);
	if(n3.size() < n2.size())
	{
		/* Swap them over */
		node::dereference(n);
		n = newparent;
	}
	else
	{
		node::dereference(newparent);
	}
#endif
}

void node::emit_call() const
{
	string name = get_name();
	fprintf(fout,"\t\t%s%s%s(PARAMS",(isvoid?"":"return "),prefix,name.c_str());
	if(function_args.size())
	{
		fprintf(fout," PARAMSCOMMA ");
		for(map<string,argument>::const_iterator it=function_args.begin();it!=function_args.end();++it)
		{
			if(it != function_args.begin())
				fprintf(fout,",");
			fprintf(fout,"%s",it->first.c_str());
		}
	}
	fprintf(fout,");\n%s",(isvoid?"\t\treturn;\n":""));
}

void node::emit_prototype()
{
	if(emitted_prototype)
		return;
	emitted_prototype = true;
	string name = get_name();
	fprintf(fout,"static %sRETURNTYPE %s%s(PARAMDECL",(((refcount==1)&&!treewalker)?"inline ":""),prefix,name.c_str());
	emit_args();
	fprintf(fout,");\n");
	/* and the children */
	for(uint32_t i=0;i<numchildren;i++)
		if(children[i])
			children[i]->emit_prototype();
}

void node::emit_args() const
{
	if(!function_args.size())
		return;
	fprintf(fout," PARAMSCOMMA ");
	for(map<string,argument>::const_iterator it=function_args.begin();it!=function_args.end();++it)
	{
		const argument &arg = it->second;
		if(it != function_args.begin())
			fprintf(fout,",");
		switch(arg.type)
		{
		case EARG_FIELD:
		case EARG_NONSTANDARD:
			fprintf(fout,"const uint32_t %s",((gettype() == ELEAF)?arg.str.c_str():it->first.c_str()));
			break;
		case EARG_ARGSTRUCT:
			fprintf(fout,"const %sargstruct *%sargs",prefix,prefix);
			break;
		default:
			assert(0);
		}
	}
}

void node::emit_locals() const
{
//	fprintf(fout,"\t/* %d refs */\n",refcount);
	for(map<string,argument>::const_iterator it=local_vars.begin();it!=local_vars.end();++it)
	{
		const argument &arg = it->second;
		switch(arg.type)
		{
		case EARG_FIELD:
			fprintf(fout,"\tconst uint32_t %s",((gettype() == ELEAF)?arg.str.c_str():it->first.c_str()));
			if(useargstruct)
				fprintf(fout," = %sargs->%s;\n",prefix,it->first.c_str());
			else
			{
				uint32_t offset = 0;
				for(size_t i=0;i<arg.fields.size();i++)
				{
					fprintf(fout," %c (((OPCODE >> %d) & 0x%x)<<%d)",(i?'|':'='),arg.fields[i]->startbit,((((uint32_t)1)<<arg.fields[i]->bitcount)-1),offset);
					offset += arg.fields[i]->bitcount;
				}
				fprintf(fout,";\n");
			}
			break;
		case EARG_NONSTANDARD:
			fprintf(fout,"\tconst uint32_t %s",((gettype() == ELEAF)?"nonstandard":it->first.c_str()));
			if(!arg.shouldmatch_mask)
				fprintf(fout," = 0;\n");
			else if(useargstruct)
				fprintf(fout," = %sargs->%s;\n",prefix,it->first.c_str());
			else
				fprintf(fout," = ((OPCODE & 0x%x) ^ 0x%x);\n",arg.shouldmatch_mask,arg.shouldmatch_value);
			break;
		default:
			assert(0);
		}
	}
}

void node::prepare_args()
{
	if(prepared_args)
		return;

	prepared_args = true;

	/* Prepare the children */
	for(uint32_t i=0;i<numchildren;i++)
		if(children[i])
			children[i]->prepare_args();
}

void node::treewalk_number(ntype type,int *count,vector<node *> &nodes)
{
	if((type == gettype()) && (treewalk_nodeid == -1))
	{
		treewalk_nodeid = (*count)++;
		nodes.push_back(this);
	}
	for(uint32_t i=0;i<numchildren;i++)
		if(children[i])
			children[i]->treewalk_number(type,count,nodes);
}

node_constraint::node_constraint(const constraint *c,const opcode_subset *o,const encodingset *res)
: node(2,o,res)
{
	con = c;
	subset = *o;
	/* constrain children */
	childset[0].constrain(c,true);
	childset[1].constrain(c,false);
	/* TODO - optimise the below or just get rid of it! */
//	childset[0].constrain(res[0]);
//	childset[1].constrain(res[1]);
}

node_constraint::~node_constraint()
{
}

ntype node_constraint::gettype() const
{
	return ECONSTRAINT;
}

void node_constraint::emit_prototype()
{
	if(emitted_prototype)
		return;
	/* base functionality */
	node::emit_prototype();
	/* and any required functions */
	con->emit_funcs();
}

void node_constraint::emit_body()
{
	if(emitted_body)
		return;
	emitted_body = true;
	string name = get_name();
	fprintf(fout,"RETURNTYPE %s%s(PARAMDECL",prefix,name.c_str());
	emit_args();
	fprintf(fout,")\n{\n");
	emit_locals();
	fprintf(fout,"\tif(");
	con->emit("OPCODE");
	fprintf(fout,")\n\t{\n");
	children[0]->emit_call();
	fprintf(fout,"\t}\n\telse\n\t{\n");
	children[1]->emit_call();
	fprintf(fout,"\t}\n}\n");
	/* and the children */
	children[0]->emit_body();
	children[1]->emit_body();
}

void node_constraint::emit_graph()
{
	if(emitted_graph)
		return;
	emitted_graph = true;
	string name = get_name();
	fprintf(fgraph,"<node id=\"%s\">\n <desc>{",name.c_str());
	string desc = con->describe(NULL,0);
	/* Output string, escaping any <'s we find */
	for(size_t i=0;i<desc.length();i++)
		if(desc[i] == '<')
			fputs("&lt;",fgraph);
		else
			fputc(desc[i],fgraph);
	fprintf(fgraph,"}</desc>\n");
#ifdef YED
	/* Output fudgey elements for yEd to parse */
	fprintf(fgraph," <data key=\"a\">\n<y:ShapeNode>\n<y:NodeLabel>{");
	for(size_t i=0;i<desc.length();i++)
		if(desc[i] == '<')
			fputs("&lt;",fgraph);
		else
			fputc(desc[i],fgraph);
	fprintf(fgraph,"}</y:NodeLabel></y:ShapeNode></data>\n");
#endif
	fprintf(fgraph,"</node>\n");
	/* Edges & children */
	for(uint32_t i=0;i<numchildren;i++)
	{
		if(children[i])
		{
			children[i]->emit_graph();
			string cname = children[i]->get_name();
			fprintf(fgraph,"<edge source=\"%s\" target=\"%s\">\n <desc>%s</desc>\n",name.c_str(),cname.c_str(),(i?"FALSE":"TRUE"));
#ifdef YED
			fprintf(fgraph," <data key=\"b\"><y:PolyLineEdge><y:EdgeLabel>%s</y:EdgeLabel></y:PolyLineEdge></data>\n",(i?"FALSE":"TRUE"));
#endif
			fprintf(fgraph,"</edge>\n");
		}
	}
}

string node_constraint::get_name() const
{
	char buf[128];
	sprintf(buf,"constraint_%x_%x_%x_%p",subset.fixed,subset.mask,con->getfields(),this); /* Use our address as a unique ID to avoid problems with duplicates (e.g. due to constraints being matched/not matched) */
	return string(buf);
}

const encoding *node_constraint::verify(uint32_t opcode) const
{
	if(con->eval(opcode))
		return children[0]->verify(opcode);
	else
		return children[1]->verify(opcode);
}

bool node_constraint::operator==(const node &n) const
{
	if(n.gettype() != ECONSTRAINT)
		return false;
	const node_constraint *nc = static_cast<const node_constraint *>(&n);
	return ((nc->con == con) && (nc->children[0] == children[0]) && (nc->children[1] == children[1]));
}

bool node_constraint::sametest(const node &n) const
{
	if(n.gettype() != ECONSTRAINT)
		return false;
	const node_constraint *nc = static_cast<const node_constraint *>(&n);
	return (nc->con == con);
}

const constraint *node_constraint::getconstraint() const
{
	return con;
}

node_mask::node_mask(uint32_t start,uint32_t count,const opcode_subset *o,const encodingset *res)
: node(((uint32_t)1)<<count,o,res)
{
	startbit = start;
	bitcount = count;
	subset = *o;
	/* Constrain children */
	uint32_t mask = (((uint32_t)1<<count)-1)<<start;
	for(uint32_t i=0;i<numchildren;i++)
	{
		childset[i].constrain(i<<start,mask);
		/* TODO - optimise the below or just get rid of it! */
//		childset[i].constrain(res[i]);
	}
}

node_mask::~node_mask()
{
}

ntype node_mask::gettype() const
{
	return EMASK;
}

void node_mask::emit_body()
{
	if(emitted_body)
		return;
	emitted_body = true;

	/* emit our children's bodies */
	for(uint32_t i=0;i<numchildren;i++)
		if(children[i])
			children[i]->emit_body();

	string name = get_name();
	fprintf(fout,"RETURNTYPE %s%s(PARAMDECL",prefix,name.c_str());
	emit_args();
	fprintf(fout,")\n{\n");
	emit_locals();
	uint32_t bits = numchildren-1;
	uint32_t mask = bits<<startbit;
	/* Decide how we want to represent this test
	   Ordinarily we'd just use a switch() statement, but to give the compiler a helping hand we'll optimise some of the cases ourselves */
	uint32_t remap[numchildren]; /* used to track duplicate children */
	uint32_t count[numchildren]; /* how many times each duplicate child occurs */
	uint32_t reduce_remap[numchildren];
	uint32_t inv_reduce_remap[numchildren];
	uint32_t cases=0; /* number of children after resolving duplicates */
	uint32_t groups=1; /* number of contiguous groups */
	uint32_t groupstart[numchildren]; /* which children[] index each group starts at */
	memset(&remap[0],0xff,sizeof(uint32_t)*numchildren);
	groupstart[0] = 0;
	uint32_t first=0;
	while(!children[first])
		first++;
	for(uint32_t i=0;i<numchildren;i++)
	{
		fprintf(fout,"\t/* 0x%x -> %p */\n",i,children[i]); /* Dump table entries for manual verification of logic */
		/* Fill in blank entries in fuzzy trees. Note that this doesn't attempt to fill them in in any way that would avoid unnecesarrily complicating the decode logic. */
		if(!children[i])
		{
			children[i] = children[first];
			children[i]->reference();
		}
		else
			first = i;
	}
	uint32_t reducecount=0;
	for(uint32_t i=0;i<numchildren;i++)
	{
		count[i] = 0;
		if(remap[i] == 0xffffffff)
		{
			for(uint32_t j=i;j<numchildren;j++)
			{
				if(children[i] == children[j])
				{
					remap[j] = i;
					reduce_remap[j] = reducecount;
					if(!(count[i]++))
						cases++;
				}
			}
			inv_reduce_remap[reducecount] = i;
			reducecount++;
		}
		if(i && (remap[i-1] != remap[i]))
		{
			groupstart[groups] = i;
			groups++;
		}
	}
	assert(groups >= cases);
	assert(cases > 1); /* Tree generation should have prevented us from having 1 (or 0!) cases? */
	if(cases == 2)
	{
		/* Candidate for if() statement */
		uint32_t a=0;
		uint32_t b=groupstart[1];
		if((count[a] == 1) && (numchildren>2))
		{
			/* simple match against one value */
			fprintf(fout,"\tif((OPCODE & 0x%x)==0x%x)\n\t{ /* %p */\n",mask,a<<startbit,children[a]);
			children[a]->emit_call();
			fprintf(fout,"\t}\n\telse\n\t{ /* %p */\n",children[b]);
			children[b]->emit_call();
			fprintf(fout,"\t}\n}\n");
			return;
		}
		if(count[b] == 1)
		{
			/* simple match against one value */
			if(numchildren==2)
			{
				/* We're only interested in one bit; make it really obvious to the compiler! */
				fprintf(fout,"\tif(OPCODE & 0x%x)\n\t{ /* %p */\n",mask,children[b]);
			}
			else
			{
				fprintf(fout,"\tif((OPCODE & 0x%x)==0x%x)\n\t{ /* %p */\n",mask,b<<startbit,children[b]);
			}
			children[b]->emit_call();
			fprintf(fout,"\t}\n\telse\n\t{ /* %p */\n",children[a]);
			children[a]->emit_call();
			fprintf(fout,"\t}\n}\n");
			return;
		}
		if(count[a] == b)
		{
			/* simple range check; first 'b' entries are 'a' */
			fprintf(fout,"\tif((OPCODE & 0x%x)>=0x%x)\n\t{ /* %p */\n",mask,b<<startbit,children[b]);
			children[b]->emit_call();
			fprintf(fout,"\t}\n\telse\n\t{ /* %p */\n",children[a]);
			children[a]->emit_call();
			fprintf(fout,"\t}\n}\n");
			return;
		}
	}
	if(groups == 3)
	{
		uint32_t a=0;
		uint32_t b=groupstart[1];
		uint32_t c=groupstart[2];
		if(c==b+1)
		{
			/* Assume the compiler is smart enough to reduce this test to the following:
			   CMP opcode,val
			   BLT a
			   BEQ b
			   B c
			*/
			fprintf(fout,"\tconst uint32_t %sbits = OPCODE & 0x%x;\n",prefix,mask);
			fprintf(fout,"\tif(%sbits < 0x%x)\n\t{ /* %p */\n",prefix,b<<startbit,children[a]);
			children[a]->emit_call();
			fprintf(fout,"\t}\n\telse if(%sbits == 0x%x)\n\t{ /* %p */\n",prefix,b<<startbit,children[b]);
			children[b]->emit_call();
			fprintf(fout,"\t}\n\telse\n\t{ /* %p */\n",children[c]);
			children[c]->emit_call();
			fprintf(fout,"\t}\n}\n");
			return;
		}
		/* Else do a slightly more complex check:
		   CMP opcode,val1
		   BLT a
		   CMP opcode,val2
		   BLT b
		   B c */
		fprintf(fout,"\tconst uint32_t %sbits = OPCODE & 0x%x;\n",prefix,mask);
		fprintf(fout,"\tif(%sbits < 0x%x)\n\t{ /* %p */\n",prefix,b<<startbit,children[a]);
		children[a]->emit_call();
		fprintf(fout,"\t}\n\telse if(%sbits < 0x%x)\n\t{ /* %p */\n",prefix,c<<startbit,children[b]);
		children[b]->emit_call();
		fprintf(fout,"\t}\n\telse\n\t{ /* %p */\n",children[c]);
		children[c]->emit_call();
		fprintf(fout,"\t}\n}\n");
		return;
	}
	if((cases < numchildren) && (cases <= 256))
	{
		/* Assume the compiler is bad at dealing with switch() statements where multiple cases map to the same outcome. Use a lookup table to eliminate duplicate cases */
		uint32_t casespow2 = cases;
		if(casespow2 & (casespow2-1))
		{
			casespow2 |= casespow2>>1;
			casespow2 |= casespow2>>2;
			casespow2 |= casespow2>>4;
			casespow2 |= casespow2>>8;
			casespow2 |= casespow2>>16;
			casespow2++;
		}
		uint32_t caseslog2 = 0;
		while((1u<<caseslog2) != casespow2)
			caseslog2++;
		if(caseslog2*numchildren <= 32)
		{
			/* Pack the lookup table into a word */
			uint32_t casesval = 0;
			for(uint32_t i=0;i<numchildren;i++)
				casesval |= reduce_remap[i]<<(caseslog2*i);
			fprintf(fout,"\tstatic const uint32_t %scases = 0x%x;\n",prefix,casesval);
			if((caseslog2-1) & caseslog2)
			{
				fprintf(fout,"\tuint32_t %scase = (%scases>>(((OPCODE & 0x%x)>>%d)*%d)) & %d;\n",prefix,prefix,bits<<startbit,startbit,caseslog2,casespow2-1);
			}
			else
			{
				uint32_t log2caseslog2 = 0;
				while((1u<<log2caseslog2) != caseslog2)
					log2caseslog2++;
				int shift = log2caseslog2-startbit;
				if(shift<0)
					fprintf(fout,"\tuint32_t %scase = (%scases>>((OPCODE & 0x%x)>>%d)) & %d;\n",prefix,prefix,bits<<startbit,-shift,casespow2-1);
				else
					fprintf(fout,"\tuint32_t %scase = (%scases>>((OPCODE & 0x%x)<<%d)) & %d;\n",prefix,prefix,bits<<startbit,shift,casespow2-1);
			}
		}
		else
		{
			/* Use a byte lookup table */
			fprintf(fout,"\tstatic const uint8_t %scases[] = {\n\t\t",prefix);
			for(uint32_t i=0;i<numchildren;i++)
				fprintf(fout,"%d%c",reduce_remap[i],(i==numchildren-1?'\n':','));
			fprintf(fout,"\t};\n\tuint8_t %scase = %scases[(OPCODE>>%d) & 0x%x];\n",prefix,prefix,startbit,bits);
		}
		fprintf(fout,"\tswitch(%scase)\n\t{\n",prefix);
		for(uint32_t i=0;i<reducecount;i++)
		{
			uint32_t child = inv_reduce_remap[i];
			if(i == reducecount-1)
				fprintf(fout,"\tdefault:\n");
			else
				fprintf(fout,"\tcase 0x%x:\n",i);
			fprintf(fout,"\t/* %p */\n",children[child]);
			children[child]->emit_call();
		}
		fprintf(fout,"\t}\n}\n");
		return;
	}
	/* Else just emit a simplified switch() statement and leave the rest to the compiler */
	fprintf(fout,"\tswitch((OPCODE>>%d) & 0x%x)\n\t{\n",startbit,bits);
	for(uint32_t i=0;i<numchildren;i++)
		if(count[i])
		{
			if(children[i] == children[numchildren-1])
				fprintf(fout,"\tdefault:\n");
			else
			{
				for(uint32_t j=0;j<numchildren;j++)
					if(remap[j] == i)
						fprintf(fout,"\tcase 0x%x:\n",j);
			}
			fprintf(fout,"\t/* %p */\n",children[i]);
			children[i]->emit_call();
		}
	fprintf(fout,"\t}\n}\n");
}

void node_mask::emit_graph()
{
	if(emitted_graph)
		return;
	emitted_graph = true;
	string name = get_name();
	fprintf(fgraph,"<node id=\"%s\">\n",name.c_str());
	fprintf(fgraph," <desc>%u:%u</desc>\n",startbit+bitcount-1,startbit);
#ifdef YED
	/* Output fudgey elements for yEd to parse */
	fprintf(fgraph," <data key=\"a\">\n<y:ShapeNode>\n<y:NodeLabel>%u:%u</y:NodeLabel></y:ShapeNode></data>\n",startbit+bitcount-1,startbit);
#endif
	fprintf(fgraph,"</node>\n");
	/* Edges & children */
	for(uint32_t i=0;i<numchildren;i++)
	{
		if(children[i])
		{
			children[i]->emit_graph();
			string cname = children[i]->get_name();
			fprintf(fgraph,"<edge source=\"%s\" target=\"%s\">\n <desc>%u</desc>\n",name.c_str(),cname.c_str(),i);
#ifdef YED
			fprintf(fgraph," <data key=\"b\"><y:PolyLineEdge><y:EdgeLabel>%u</y:EdgeLabel></y:PolyLineEdge></data>\n",i);
#endif
			fprintf(fgraph,"</edge>\n");
		}
	}
}

string node_mask::get_name() const
{
	char buf[128];
	sprintf(buf,"mask_%x_%x_%d_%d_%p",subset.fixed,subset.mask,startbit,bitcount,this); /* Use our address as a unique ID to avoid problems with duplicates (e.g. due to constraints being matched/not matched) */
	return string(buf);
}

const encoding *node_mask::verify(uint32_t opcode) const
{
	return children[(opcode>>startbit) & ((((uint32_t)1)<<bitcount)-1)]->verify(opcode);
}

bool node_mask::operator==(const node &n) const
{
	if(n.gettype() != EMASK)
		return false;
	const node_mask *nm = static_cast<const node_mask *>(&n);
	if((nm->startbit != startbit) || (nm->bitcount != bitcount))
		return false;
	for(uint32_t i=0;i<numchildren;i++)
		if(nm->children[i] != children[i])
			return false;
	return true;
}

bool node_mask::sametest(const node &n) const
{
	if(n.gettype() != EMASK)
		return false;
	const node_mask *nm = static_cast<const node_mask *>(&n);
	if((nm->startbit != startbit) || (nm->bitcount != bitcount))
		return false;
	return true;
}

void node_mask::shrink2()
{
	/* Shrink the mask area to a minimum */
	while(!memcmp(children,children+(numchildren>>1),sizeof(node *)*(numchildren>>1)))
	{
		/* Topmost bit can be discarded */
		bitcount--;
		for(numchildren=0;numchildren<(((uint32_t)1)<<bitcount);numchildren++)
			if(children[numchildren])
				dereference(children[numchildren]);
	}
	do {
		for(uint32_t i=0;i<numchildren;i+=2)
			if(children[i] != children[i+1])
				return;
		/* Bottommost bit can be discarded */
		bitcount--;
		startbit++;
		numchildren >>= 1;
		for(uint32_t i=0;i<numchildren;i++)
		{
			children[i] = children[i<<1];
			if(children[i])
				dereference(children[i]);
		}
	} while(1);
}

uint8_t node_mask::getstartbit() const
{
	return startbit;
}

uint8_t node_mask::getbitcount() const
{
	return bitcount;
}

node_leaf::node_leaf(const encoding *e)
: node(0,NULL,NULL)
{
	enc = e;
	opt_action = NULL;
}

node_leaf::~node_leaf()
{
}

ntype node_leaf::gettype() const
{
	return ELEAF;
}

void node_leaf::emit_body()
{
	if(emitted_body)
		return;
	emitted_body = true;
	if(opt_action)
	{
		opt_action->emit(this);
	}
	const encoding *e = get_encoding();
	string name = get_name();
	fprintf(fout,"RETURNTYPE %s%s(PARAMDECL",prefix,name.c_str());
	emit_args();
	fprintf(fout,")\n{ /* ");
	e->emit_comment();
	fprintf(fout," */\n");
	if(opt_action)
	{
		/* Emit call to the optimised action */
		fprintf(fout,"\t%s%s%s(PARAMS PARAMSCOMMA ",(isvoid?"":"return "),prefix,opt_action->get_name().c_str());
		for(map<string,argument>::iterator it=function_args.begin();it!=function_args.end();++it)
		{
			argument &arg = it->second;
			fprintf(fout,"%s,",arg.str.c_str());
		}
		for(size_t i=0;i<opt_action->args.size();i++)
		{
		
			fprintf(fout,"%s%c",opt_action_args[opt_action->args[i]].c_str(),(i==opt_action->args.size()-1)?')':',');
		}
		fprintf(fout,";\n%s}\n",(isvoid?"\treturn;\n":""));
	}
	else
	{
		emit_locals();
		fputs(e->action->code,fout);
		fprintf(fout,"}\n");
	}
}

string node_leaf::get_name() const
{
	/* Name ourselves by our root kin, and any shouldmatch bits */
	const encoding *e = get_encoding();
	char buf[strlen(e->name)+32];
	sprintf(buf,"leaf_%s_%p",e->name,this);
	return string(buf);
}

void node_leaf::emit_graph()
{
	if(emitted_graph)
		return;
	emitted_graph = true;
	string name = get_name();
	fprintf(fgraph,"<node id=\"%s\">\n",name.c_str());
	const encoding *e = get_encoding();
	fprintf(fgraph," <desc>%s</desc>\n",e->name); /* Make this list all the kin, similar to the comments in the output source file? */
#ifdef YED
	/* Output fudgey elements for yEd to parse */
	fprintf(fgraph," <data key=\"a\">\n<y:ShapeNode>\n<y:NodeLabel>%s</y:NodeLabel></y:ShapeNode></data>\n",e->name);
#endif
	fprintf(fgraph,"</node>\n");
}

const encoding *node_leaf::verify(uint32_t opcode) const
{
	return enc;
}

bool node_leaf::operator==(const node &n) const
{
	if(n.gettype() != ELEAF)
		return false;
	const node_leaf *nl = static_cast<const node_leaf *>(&n);
	return ((nl->enc->kin?nl->enc->kin:nl->enc) == (enc->kin?enc->kin:enc)); /* Use root kin as indication of whether the nodes are identical */
}

bool node_leaf::sametest(const node &n) const
{
	/* This doesn't make much sense, since we don't test anything! */
	if(n.gettype() != ELEAF)
		return false;
	const node_leaf *nl = static_cast<const node_leaf *>(&n);
	return ((nl->enc->kin?nl->enc->kin:nl->enc) == (enc->kin?enc->kin:enc)); /* Use root kin as indication of whether the nodes are identical */
}

const encoding *node_leaf::get_encoding() const
{
	return (enc->kin?enc->kin:enc);
}

void node_leaf::set_opt_strings_action(opt_strings_action *a,map<string,string> &args)
{
	opt_action = a;
	opt_action_args = args;
}

void node_leaf::prepare_args()
{
	if(prepared_args)
		return;

	/* Create local argument list from the action */
	const encoding *e = get_encoding();
	char buf[strlen(prefix)+256];
	const actiondef *act = e->action;
	for(uint32_t i=0;i<act->numparam;i++)
	{
		const char *param = act->params[i];

		argument a;
		a.type = EARG_FIELD;
		strcpy(buf,param);
		char *colon = strchr(buf,':');
		while(colon)
		{
			*colon = '_';
			colon = strchr(colon,':');
		}
		a.str = buf;

		sprintf(buf,"%sargf",prefix);
		while(param)
		{
			const char *terminator = strchr(param,':');
			size_t len;
			if(terminator)
				len = (terminator++)-param;
			else
				len = strlen(param);
			uint32_t j;
			for(j=0;j<e->numfields;j++)
				if((strlen(e->fields[j].name) == len) && !strncmp(param,e->fields[j].name,len))
				{
					a.fields.insert(a.fields.begin(),&e->fields[j]);
					sprintf(buf+strlen(buf),"_%d_%d",e->fields[j].startbit,e->fields[j].bitcount);
					break;
				}
			param = terminator;
		}
		local_vars[buf] = a;
	}
	if(act->nonstandard)
	{
		argument a;
		a.type = EARG_NONSTANDARD;
		a.shouldmatch_mask = e->shouldmatch_mask;
		a.shouldmatch_value = e->shouldmatch_value;
		a.str = "nonstandard";
		sprintf(buf,"%sargn_%x_%x",prefix,a.shouldmatch_mask,a.shouldmatch_value);
		local_vars[buf] = a;
	}

	node::prepare_args();
}
